//==========================================================================
//
// mc9328mxl_mem.h
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         07/18/2003
// Description:  This file contains register base addresses and offsets
//				 and access macros for the MC9328MXL memory controller
//

#include "bits.h"


// EIM - External Interface Module
#define EIM_CS0U 			0x00 	// Chip Select 0 Upper Control Register
#define EIM_CS0L 			0x04 	// Chip Select 0 Lower Control Register
#define EIM_CS1U 			0x08 	// Chip Select 1 Upper Control Register
#define EIM_CS1L 			0x0C 	// Chip Select 1 Lower Control Register
#define EIM_CS2U 			0x10 	// Chip Select 2 Upper Control Register
#define EIM_CS2L 			0x14 	// Chip Select 2 Lower Control Register
#define EIM_CS3U 			0x18 	// Chip Select 3 Upper Control Register
#define EIM_CS3L 			0x1C 	// Chip Select 3 Lower Control Register
#define EIM_CS4U 			0x20 	// Chip Select 4 Upper Control Register
#define EIM_CS4L 			0x24 	// Chip Select 4 Lower Control Register
#define EIM_CS5U 			0x28 	// Chip Select 5 Upper Control Register
#define EIM_CS5L 			0x2C 	// Chip Select 5 Lower Control Register
#define EIM_CFG				0x30 	// Configuration Register


// Bit Defines
// EIM_CS0U to EIM_CS5U - Chip Select Upper Control Register
#define EIM_CSU_DTACK_SEL	BIT31					// 0 = Generic DTACK function, 1 = PCMCIA wait function
#define EIM_CSU_BCD(_x_)	((_x_ & 0x03) << 28)	// Burst Clock Divisor, when EIM_CFG_BCM = 0
#define EIM_CSU_BCS(_x_)	((_x_ & 0x03) << 24)	// Burst Clock Start, # of 1/2 cycles from LBA to BCLK high
#define EIM_CSU_PSZ_4		(0 << 22)				// page size = 4 words
#define EIM_CSU_PSZ_8		(1 << 22)				// page size = 8 words
#define EIM_CSU_PSZ_16		(2 << 22)				// page size = 16 words
#define EIM_CSU_PSZ_32		(3 << 22)				// page size = 32 words
#define EIM_CSU_PME			BIT21					// 1 = Enables page mode emulation
#define EIM_CSU_SYNC		BIT20					// 1 = Enables synchronous burst mode
#define EIM_CSU_DOL(_x_)	((_x_ & 0x0f) << 16)	// # of clocks -1 before latching read data when SYNC = 1		
#define EIM_CSU_CNC_0		(0 << 14)				// Hold CS negated after end of cycle for 0 clocks
#define EIM_CSU_CNC_1		(1 << 14)				// Hold CS negated after end of cycle for 1 clock
#define EIM_CSU_CNC_2		(2 << 14)				// Hold CS negated after end of cycle for 2 clocks
#define EIM_CSU_CNC_3		(3 << 14)				// Hold CS negated after end of cycle for 3 clocks
#define EIM_CSU_WSC(_x_)	((_x_ & 0x3f) << 8)		// Wait States, 0 = 2, 1 = 2, 2-62 = +1, 63 = dtack
#define EIM_CSU_WSC_DTACK	(0x3f << 8)		        // Wait States, 0 = 2, 1 = 2, 2-62 = +1, 63 = dtack
#define EIM_CSU_WWS(_x_)	((_x_ & 0x07) << 4)		// Additional wait states for write cycles
#define EIM_CSU_EDC(_x_)	((_x_ & 0x0f) << 0)		// Dead Cycles after reads for bus turn around

// EIM_CS0L to EIM_CS5L - Chip Select Lower Control Register
#define EIM_CSL_OEA(_x_)	((_x_ & 0x0f) << 28)	// # of 1/2 cycles after CS asserts before OE asserts
#define EIM_CSL_OEN(_x_)	((_x_ & 0x0f) << 24)	// # of 1/2 cycles OE negates before CS negates
#define EIM_CSL_WEA(_x_)	((_x_ & 0x0f) << 20)	// # of 1/2 cycles EB0-3 assert before WE asserts
#define EIM_CSL_WEN(_x_)	((_x_ & 0x0f) << 16)	// # of 1/2 cycles EB0-3 negate before WE negates 
#define EIM_CSL_CSA(_x_)	((_x_ & 0x0f) << 12)	// # of clocks address is asserted before CS and held after CS
#define EIM_CSL_EBC			BIT11					// 0 = assert EB0-3 for Reads & Writes, 1 = Writes only
#define EIM_CSL_DSZ_BYTE_3	(0 << 8)				// device is 8-bits wide, located on byte 3 (D31-24)
#define EIM_CSL_DSZ_BYTE_2	(1 << 8)				// device is 8-bits wide, located on byte 2 (D23-16)
#define EIM_CSL_DSZ_BYTE_1	(2 << 8)				// device is 8-bits wide, located on byte 1 (D15-8)
#define EIM_CSL_DSZ_BYTE_0	(3 << 8)				// device is 8-bits wide, located on byte 0 (D7-0)
#define EIM_CSL_DSZ_HALF_1	(4 << 8)				// device is 16-bits wide, located on half word 1 (D31-16)
#define EIM_CSL_DSZ_HALF_0	(5 << 8)				// device is 16-bits wide, located on half word 1 (D15-0)
#define EIM_CSL_DSZ_WORD	(6 << 8)				// device is 32-bits wide, located on half word 1 (D31-0)
#define EIM_CSL_SP			BIT6					// 0 = User mode access allowed, 1= Supervisor mode only
#define EIM_CSL_WP			BIT4					// 0 = Writes allowed, 1 = Write Protected
#define EIM_CSL_PA			BIT1					// Set the pin to EIM_CSL_PA value when CSEN is 0
#define EIM_CSL_CSEN		BIT0					// 1 = Enable Chip Select

// EIM_CFG - Configuration Register
#define EIM_CFG_BCM			BIT2					// 1 = Burst Clock always on, 0 = when CS with SYNC = 1 is accessed


// ------------------------------------------
// SDRAMC - SDRAM Controller
#define SDRAMC_CTL0 		0x00 	// SDRAM 0 Control Register
#define SDRAMC_CTL1 		0x04 	// SDRAM 1 Control Register
#define SDRAMC_MISC 		0x14 	// Miscellaneous Register
#define SDRAMC_RST 			0x18 	// SDRAM Reset Register
