//==========================================================================
//
//      sm501_vid.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        05-27-2002
// Description:	Init Code for sm501 VID Controller
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "sm501.h"
#include "sm501_video.h"
#include "font8x16.h"
#include "vid_lut.h"

//#define VID_DBG

//--------------------------------------------------------------------------
// The Video frame buffer is fixed at 0xa1000000, which is 1Mbyte from the
// beginning of SDRAM space.  Note that we access it 16-bits at a time.
//
#define VID_PDR_ADD		0xa0100000
#define VID_PDR(_x_)	*(vulong *)(VID_PDR_ADD + _x_)	// Palette Descriptor, Palette Buffer

#define VID_FDR_ADD		0xa0100010
#define VID_FDR(_x_)	*(vulong *)(VID_FDR_ADD + _x_)	// Frame Descriptor, Frame Buffer

#define VID_PAL_ADD		0xa0100020
#define VID_PAL(_x_)	*(vushort *)(VID_PAL_ADD + _x_)	// Palette Buffer

#define VID_BUF_ADD		0xa0100400
#define VID_BUF(_x_)	*(vushort *)(VID_BUF_ADD + _x_)	// Frame Buffer

//--------------------------------------------------------------------------
// function prototypes and externs
//
int vid_init(void);
void vid_on(void);
void vid_off(void);

void vid_putchar(char c);
void vid_writechar(uchar c);
void vid_clr_row(int char_row);

extern void udelay(int);
extern int GPIO_set(int);
extern int GPIO_clr(int);
extern int GPIO_tst(int);

// globals to keep track of foreground, background colors and x,y position
int vid_color_depth;		// 4, 8 or 16
int vid_fg_color;			// 0 to 15, used as lookup into VGA color table
int vid_bg_color;			// 0 to 15, used as lookup into VGA color table
int vid_col;				// current column, 0 to COLS_PER_SCREEN - 1
int vid_row;				// current row, 0 to (ROWS_PER_SCREEN * 2) - 1
int vid_tst_mode = 0;
ulong vid_fb_offset;	   	// current offset into frame buffer for vid_putchar

#define PIXELS_PER_ROW			640
#define PIXELS_PER_COL			480

#define ROWS_PER_SCREEN			30
#define COLS_PER_SCREEN			80

#define VID_BG_DEF				1
#define VID_FG_DEF				14

#define VID_FB_SIZE(_depth_)				(((PIXELS_PER_COL * PIXELS_PER_ROW) * _depth_) / 8)

#define VID_ROW_SIZE(_depth_)				((PIXELS_PER_ROW * _depth_) / 8)

#define VID_GET_ADD(_row_, _col_, _depth_)	(((((_row_ * PIXELS_PER_ROW) * FONT_HEIGHT) \
											  + (_col_ * FONT_WIDTH)) \
											  * _depth_) / 8)

//--------------------------------------------------------------------------
// vid_init
//
// This function sets up the sm501 VID controller
//
int vid_init()
{
	ulong index;
	int i;
	ushort temp16;

#ifdef VID_DBG
	vid_tst_mode = 1;
	printf("sm501 VID Initialization.\n");
	vid_tst_mode = 0;
#endif

	// Setup Timing Registers
	// Note, VID pixel clock is derived from MEMCLK which
	// defaults to 99.5Mhz.  Any other value requires changing
	// LCCR3_PCD.
	VID_REG(VID_LCCR3) = LCCR3_PCP								// Pixel Clock Polarity, 1 = VID Data is Sampled on Falling Edge
					   | LCCR3_HSP								// Horizontal Sync Polarity, 1 = Active Low
					   | LCCR3_VSP								// Vertical Sync Polarity, 1 = Active Low
				       | LCCR3_PCD(1);							// Pixel Clock Divisor, PCLK = LCLK/(2*(PCD+1)) = 99.5Mhz/4 = 25Mhz

	// set the bpp based on vid_color_depth
	switch (vid_color_depth)
	{
		case 4:
			VID_REG(VID_LCCR3) |= LCCR3_BPP_4;					// 4-bits/pixel
			break;
		case 8:
			VID_REG(VID_LCCR3) |= LCCR3_BPP_8;					// 8-bits/pixel
			break;
		default:
			VID_REG(VID_LCCR3) |= LCCR3_BPP_16;					// 16-bits/pixel
			break;
	} // switch color depth

	VID_REG(VID_LCCR2) = LCCR2_BFW(0)							// Beginning-of-frame line clock wait count
					   | LCCR2_EFW(31)							// End-of-frame line clock wait count
					   | LCCR2_VSW(31)							// Vertical sync pulse width = VSW + 1
					   | LCCR2_LPP(PIXELS_PER_COL - 1);			// Lines per panel = LPP + 1

	VID_REG(VID_LCCR1) = LCCR1_BLW(0)  							// Beginning-of-Line pixel clock Wait count = BLW + 1
	 		    	   | LCCR1_ELW(160)  						// End-of-line pixel clock wait count = ELW + 1
	 		    	   | LCCR1_HSW(0)  							// Horizontal sync pulse width = HSW + 1
			 		   | LCCR1_PPL(PIXELS_PER_ROW - 1);		   	// Pixels per line = PPL + 1


	index = 0;
	// Fill the pallette buffer with the lookup table data
	for (i = 0; i < 256; i ++)
	{
		VID_PAL(index) = (((vid_lut[i][0] & 0xf8) << 8)		// Red
					    | ((vid_lut[i][1] & 0xfc) << 3) 	// Green
					    | ((vid_lut[i][2]) >> 3));			// Blue

		index += 2;
	}

	// Setup the final VID control register, but
	// leave the controller off.  It will be controlled 
	// by vid_on() and vid_off() which also handles the
	// backlight
	VID_REG(VID_LCCR0) = LCCR0_OUM		// Output FIFO Underrun Mask, 1 = no interrupt
					   | LCCR0_BM 		// Branch Mask, 1 = no interrupt
					   | LCCR0_QDM		// Quick Disable Interrupt Mask, 1 = no interrupt
					   | LCCR0_PAS		// Passive/Active Display select 1 = TFT
					   | LCCR0_EFM		// End of Frame Interrupt Mask, 1 = no interrupt
					   | LCCR0_IUM		// Input FIFO Underrun Mask, 1 = no interrupt
					   | LCCR0_SFM		// Start of Frame Mask, 1 = no interrupt
					   | LCCR0_LDM;		// VID Disable Done Mask, 1 = no interrupt


	// initialize the globals	
	vid_fg_color = VID_FG_DEF;
	vid_bg_color = VID_BG_DEF;
	vid_row = 0;
	vid_col = 0;
	vid_fb_offset = 0x00;

	// fill the screen with the background color, remember for the console driver
	// we do double buffering, so fill twice as mush as the active screen size
	switch (vid_color_depth){
		case 4:  temp16 = vid_bg_color | vid_bg_color << 4 | vid_bg_color << 8 | vid_bg_color << 12; break;
		case 8:  temp16 = vid_bg_color | vid_bg_color << 8; break;
		default: temp16 = vga_lookup[vid_bg_color]; break;
	}
	for (i = 0; i < (PIXELS_PER_COL * PIXELS_PER_ROW) * 2; i += 2){	
		VID_BUF(i) = temp16;
	}

	return 0;
}

//--------------------------------------------------------------------------
// vid_on
//
// This function turns on the sm501 VID Controller and the Backlight Inverter
//
void vid_on()
{
	// Make sure it's off first
//	vid_off();

	// kick off the frame descriptors
	// Setup the Palette Descriptor
	VID_PDR(0x0) = VID_FDR_ADD;								// point to the frame buffer descriptor
	VID_PDR(0x4) = VID_PAL_ADD;								// beginning of Palette Buffer in Memory
	VID_PDR(0x8) = 0;										// Frame ID
	VID_PDR(0xc) = BIT26 | 512;								// PAL = 1, Frame Length in bytes = 512

	// Setup the Frame Descriptor
	VID_FDR(0x0) = VID_FDR_ADD;								// point it to itself
	VID_FDR(0x4) = VID_BUF_ADD;								// beginning of Frame Buffer in Memory
	VID_FDR(0x8) = 1;										// Frame ID
	VID_FDR(0xc) = VID_FB_SIZE(vid_color_depth);			// Frame Length in bytes

	// For color depths of 4 or 8 we point the sm501 to 
	// the palette buffer descriptor, which will then point
	// to the frame buffer descriptor, while for 16 bpp 
	// we point directly to the frame buffer descriptor,
	// thus skipping the palette buffer descriptor.
	if ((vid_color_depth == 4) || (vid_color_depth == 8)){
//		VID_FDR(0x0) = VID_PDR_ADD;								// point back to the palette descriptor 
		VID_REG(VID_FDADR0) = VID_PDR_ADD;
	}
	else {
//		VID_FDR(0x0) = VID_FDR_ADD;								// point it to itself as it's the only descriptor
		VID_REG(VID_FDADR0) = VID_FDR_ADD;
	}

	// Turn on the VID Controller
	VID_REG(VID_LCCR0) |= (VID_REG(VID_LCCR0) | LCCR0_ENB); 

	// Enable the Backlight
	GPIO_set(VID_BKL);

	// wait for power-up sequence to complete
	udelay(10000);
}

//--------------------------------------------------------------------------
// vid_off
//
// This function turns off the VID Controller and the VID backlight
//
void vid_off()
{
	// Turn off the VID Controller using the diable bit
	VID_REG(VID_LCCR0) |= (VID_REG(VID_LCCR0) | LCCR0_DIS);
	
	// Wait for the Enable Bit to go clear
	while(VID_REG(VID_LCCR0) & LCCR0_ENB) {}

	// Turn off the backlight
	GPIO_clr(VID_BKL);
}

char *vid_tstHelp[] = {
	" This command allows the user to test ",
	" the sm501 VID Controller.  The user ",
	" may set the color depth to run the ",
	" test at.  Note that the frame buffer ",
	" R/W test will test all of the frame ",
	" buffer regardless of depth.\n",
    " Usage:",
    " sed_tst -[n,x,d:[4,8,16]]",
    " Options...",
    " -n run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	" -d4 run test, force a depth of 4-bits/pixel",
	" -d8 run test, force a depth of 8-bits/pixel",
	" -d16 run test, force a depth of 16-bits/pixel",
	" -x init only, do not run frame buffer tests",
	" no options, default to current mode and depth",
	0
};

int vid_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16, temp16;
	int i, x, opt;
	ulong vid_fb_end;
	int no_wait = 0;
	int init_only = 0;
	char c;

	vid_tst_mode = 1;
	vid_off();

    while ((opt=getopt(argc,argv,"clnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        vid_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        vid_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    vid_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing VID, Skipping testsp!\n");
				init_only = 1;
				break;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the VID controller
	vid_init();
	vid_on();

	if (init_only) return 0;

	printf("Frame Buffer R/W...");
	// do an address=data read/write test on the frame buffer
	// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
	// Multiply by bits_per_pixel (sed_color_depth), then
	// divide by 8 to get the actual byte count.
	for (i = 0; i < VID_FB_SIZE(vid_color_depth); i += 2){	
		VID_BUF(i) = i & 0xffff;
		rd16 = VID_BUF(i);
		if(rd16 != (i & 0xffff)){
			printf("Fail at 0x%08x, WR 0x%08lx, RD 0x%04x!\n",VID_BUF_ADD + i, i, rd16);
			return -1;
		}
	} 

	printf("OK!, Press key to continue.\n");

	c = getchar();

	printf("Frame Buffer Start: 0x%08lx, End 0x%08lx\n",VID_BUF_ADD, VID_BUF_ADD + VID_FB_SIZE(vid_color_depth));
	if (no_wait)
	{
		printf("Begin Full Screen Color Test.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (vid_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < VID_FB_SIZE(vid_color_depth); i += 2){	
					VID_BUF(i) = wr16;
				}
			} // for x
		} // while
	} // no_wait
	else
	{
		printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (vid_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < VID_FB_SIZE(vid_color_depth); i += 2){	
					VID_BUF(i) = wr16;
				}
				c = getchar();
				if (c == 'x') goto vid_tst_next;
			} // for x
		} // while
	} // else no keycheck test

	vid_tst_next:

	vid_off();
	vid_init();
	vid_on();
	vid_tst_mode = 0;
	return 0;
}

// vid_scroll()
//
// Because we are most likely running out of FLASH and probably also with
// cache disabled, a brute force memcpy of the whole screen would be very
// slow, even with reduced color depths.  Instead, we define a frame buffer
// that is twice the size of our actual display.
//
// With the double buffering, we always write to the first buffer, even
// when the second buffer is active.  This allows us to scroll by adjusting
// the starting and ending addresses in the sm501 by one row.  When we
// reach the end of our virtual buffer, we reset the starting and ending
// addresses to the first buffer.
//
ulong vid_scroll(){

	ushort temp16;

	vid_row++;

	// clear the new row(s)
	vid_clr_row(vid_row);
	if (vid_row > (ROWS_PER_SCREEN - 1)){
		vid_clr_row(vid_row - ROWS_PER_SCREEN);
	}
	// when vid_y_pos is greater than ROWS_PER_SCREEN we just adjust the
	// start and end addresses in the sm501.  If it is equal to 2 *
	// ROWS_PER_SCREEN, we reset the start and end addresses to VID_MEM_BASE.
	if (vid_row > (ROWS_PER_SCREEN - 1))
	{
		if (vid_row > ((ROWS_PER_SCREEN * 2) - 1))
		{
			vid_fb_offset = 0x00;
			vid_row = ROWS_PER_SCREEN - 1;
		}
		else
		{
			// calculate the new offset address of the frame buffer in words
			vid_fb_offset += (VID_GET_ADD(1, 0, vid_color_depth));
		}

#ifdef VID_DBG
	vid_tst_mode = 1;
	printf("VID New FB Offset = 0x%08lx.\n ", vid_fb_offset);
	vid_tst_mode = 0;
#endif
		VID_FDR(0x4) = VID_BUF_ADD + vid_fb_offset;				// new beginning of Frame Buffer in Memory

	} // if (vid_row > (ROWS_PER_SCREEN - 1))
	return;
		

} // 

//--------------------------------------------------------------------------
// vid_clr_row()
//
// This routine writes the background color to the font row specified
//
void vid_clr_row(int char_row){

	ulong vid_mem_add;
	int i;
	ushort wr16;

	// clear the desired row
	vid_mem_add = VID_GET_ADD(char_row, 0, vid_color_depth);

#ifdef VID_DBG
	vid_tst_mode = 1;
	printf("VID Clear Row %d, FB Add 0x%08lx.\n ", char_row, vid_mem_add);
	vid_tst_mode = 0;
#endif

	switch (vid_color_depth)
	{
		case 4:
			wr16 = ((vid_bg_color << 12) | (vid_bg_color << 8) | (vid_bg_color << 4) | (vid_bg_color << 0));
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) / 2); i += 2){
				VID_BUF(vid_mem_add) = wr16;
				vid_mem_add += 2;
			} // for font_row
			break;
		case 8:
			wr16 = ((vid_bg_color << 8) | (vid_bg_color << 0));
			for (i = 0; i < (PIXELS_PER_ROW * FONT_HEIGHT); i += 2){
				VID_BUF(vid_mem_add) = wr16;
				vid_mem_add += 2;
			} // for font_row
			break;
		case 16:
			wr16 = (vga_lookup[vid_bg_color] << 0);
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) * 2); i += 2){
				VID_BUF(vid_mem_add) = wr16;
				vid_mem_add += 2;
			} // for font_row
			break;
	} // switch vid_color_depth
} // vid_clr_row


//--------------------------------------------------------------------------
// vid_putchar()
//
// This routine parses the character and calls vid_writechar if it is a
// printable character
//
void vid_putchar(char c){

	if (vid_tst_mode) return;

	// First parse the character to see if it is printable or an acceptable control
    switch (c) {
	    case '\r':
	        vid_col = 0;
	        return;
	    case '\n':
	        vid_scroll();
	        return;
	    case '\b':
	        vid_col--;
	        if (vid_col < 0)
	        {
	            vid_row--;
	            if (vid_row < 0) vid_row = 0;
	            vid_col = COLS_PER_SCREEN - 1;
		    }
			c = 0;		// erase the character
			vid_writechar(c);
	        break;
	    default:
	        if (((uchar)c < FIRST_CHAR) || ((uchar)c > LAST_CHAR)) return; // drop anything we can't print
			c -= FIRST_CHAR;	// get aligned to the first printable character
			vid_writechar(c);
			// advance to next column
			vid_col++;
		    if (vid_col == COLS_PER_SCREEN)
		    {
				vid_col = 0;
			    vid_scroll();
			}
			break;
    } 

} // vid_putchar()

//--------------------------------------------------------------------------
// vid_writechar()
//
// This routine writes the character to the screen at the current cursor
// location.
//
void vid_writechar(uchar c) {

	ulong vid_mem_add;
	int font_row, font_col;
	uchar font_data8;
	ushort font_data16, wr16;

	// Convert the current row,col and color depth values 
	// into an address
	vid_mem_add = VID_GET_ADD(vid_row, vid_col, vid_color_depth);

#ifdef VID_DBG
	vid_tst_mode = 1;
	printf("SED writechar at row %d, col %d, FB Add 0x%08lx.\n ", vid_row, vid_col, vid_mem_add);
	vid_tst_mode = 0;
#endif

	if (FONT_WIDTH == 8)
	{
		switch (vid_color_depth){
			case 4:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = font8x16[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 4)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? vid_fg_color << 0 : vid_bg_color << 0)
							  | ((font_data8 & 0x40) ? vid_fg_color << 4 : vid_bg_color << 4)
							  | ((font_data8 & 0x20) ? vid_fg_color << 8 : vid_bg_color << 8)
							  | ((font_data8 & 0x10) ? vid_fg_color << 12 : vid_bg_color << 12));
						font_data8 = font_data8 << 4;
//						wr16 = (((font_data8 & 0x80) ? vid_fg_color << 12 : vid_bg_color << 12)
//							  | ((font_data8 & 0x40) ? vid_fg_color << 8 : vid_bg_color << 8)
//							  | ((font_data8 & 0x20) ? vid_fg_color << 4 : vid_bg_color << 4)
//							  | ((font_data8 & 0x10) ? vid_fg_color << 0 : vid_bg_color << 0));
//						font_data8 = font_data8 << 4;
						VID_BUF(vid_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (vid_row > (ROWS_PER_SCREEN - 1)){
							VID_BUF(vid_mem_add - VID_FB_SIZE(vid_color_depth)) = wr16;
						}
						vid_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					vid_mem_add += (VID_ROW_SIZE(vid_color_depth) - ((FONT_WIDTH * vid_color_depth) / 8));
				} // for font_row
				break;

			case 8:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = font8x16[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 2)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? vid_fg_color << 0 : vid_bg_color << 0)
							  | ((font_data8 & 0x40) ? vid_fg_color << 8 : vid_bg_color << 8));
						font_data8 = font_data8 << 2;
//						wr16 = (((font_data8 & 0x80) ? vid_fg_color << 8 : vid_bg_color << 8)
//							  | ((font_data8 & 0x40) ? vid_fg_color << 0 : vid_bg_color << 0));
//						font_data8 = font_data8 << 2;
						VID_BUF(vid_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (vid_row > (ROWS_PER_SCREEN - 1)){
							VID_BUF(vid_mem_add - VID_FB_SIZE(vid_color_depth)) = wr16;
						}
						vid_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					vid_mem_add += (VID_ROW_SIZE(vid_color_depth) - ((FONT_WIDTH * vid_color_depth) / 8));
				} // for font_row
				break;

			default:	// 16bpp is the default
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = font8x16[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col++)
					{
						// get a words worth of pixels
						wr16 = ((font_data8 & 0x80) ? vga_lookup[vid_fg_color] : vga_lookup[vid_bg_color]);
						font_data8 = font_data8 << 1;
						VID_BUF(vid_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (vid_row > (ROWS_PER_SCREEN - 1)){
							VID_BUF(vid_mem_add - VID_FB_SIZE(vid_color_depth)) = wr16;
						}
						vid_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					vid_mem_add += (VID_ROW_SIZE(vid_color_depth) - ((FONT_WIDTH * vid_color_depth) / 8));
				} // for font_row
				break;
		} // switch vid_color depth
	} // FONT_WIDTH == 8
	else
	{
		return;
	}
} // vid_writechar()	

