/* low-level-driver code for the CSB536FS's compact flash.
 * Note that this code requires a modification of the delivered
 * CSB536FS.  The CF_DEN pin must be pulled low.
 * Graham Henderson (author of this driver) did found this problem
 * and worked around it by white-wiring the board.
 * I (esutter) later spoke with Cogent and confirmed the fix through
 * the CPLD.
 * Bottom line... If your CF doesn't work, then either add white wire so
 * that CF_DEN is low or update to latest CPLD (from Cogent).
 */
#include "config.h"
#include "genlib.h"
#include "cpu_gpio.h"
#include "dosfs.h"
#include "cf.h"
#include "cf_lld.h"

// ----------------------------------------------------------------------------
// wait_for_idle
//
// Wait until any outstanding IDE operations have completed. This will return
// 1 for success, or zero in the event of a timeout.
static inline int wait_for_idle( void)
{
  unsigned char status;
  unsigned int  tries;

  for (tries=0; tries < 1000000; tries++)
  {
    CF_STATUS_RD( status);

    if (!(status & IDE_STAT_BSY))
    {
      return 1;
    }
  }
  return 0;
}

// ----------------------------------------------------------------------------
// wait_for_drq
//
// Wait until the device reports that data is ready. This will return 1 for
// success, or zero in the event of a timeout.

static inline int wait_for_drq( void)
{
  unsigned char status;
  unsigned int  tries;

  int i = 0;

  for( i = 0; i < 2000; i++){}

  for (tries=0; tries < 1000000; tries++)
  {
    CF_STATUS_RD( status);

    if (!(status & IDE_STAT_BSY))
    {
      if (status & IDE_STAT_DRQ)
  		return 1;
      else
      {
  		return 0;
      }
    }
  }
  return 0;
}



int cfPresent(void)
{
  if ((GPIO_tst(CF_CD) == 0))
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


static void id_strcpy(char *dest, unsigned short *src, unsigned short size)
{
  int i;

  for (i = 0; i < size; i+=2)
  {
    *dest++ = (char)(*src >> 8);
    *dest++ = (char)(*src & 0x00FF);
    src++;
  }
  *dest = '\0';
}


static void cfReset(void)
{

    GPIO_set(CF_RST);
    monDelay(2);

    GPIO_set(CF_VEN);   //enable power
    monDelay(5);
    
    GPIO_clr(CF_RST);   // release reset
    monDelay(1000);
}

static int ide_ident( int dev,unsigned short *buf)
{
  int i;

  wait_for_drq();

  CF_REG_WR( IDE_REG_DEVICE , dev << 4);
  CF_REG_WR( IDE_REG_COMMAND, IdentifyDeviceCmd);

  monDelay(5);

  if (!wait_for_drq())
  {
    return 0;
  }

  for (i = 0; i < (512 / sizeof( unsigned short)); i++, buf++)
  {
    CF_REG_RD16( IDE_REG_DATA, *buf);
  }

  return 1;
}




int cfInit( int interface, int verbosity)
{
  unsigned int         id_buf[ 512 / sizeof(unsigned int)];
  ide_identify_disk_t  ident;
  ide_identify_data_t *ide_idData = (ide_identify_data_t*)id_buf;
  int                  rval       = 0;

  // make sure the device is physically present
  if (!cfPresent())
  {
    if( verbosity != 0) printf("CF Device(0x%x:%d) Not Found\n", CF_COMMON_MEM_BASE, interface);
    return -1;
  }

  // if verbosity enabled, print out interface and device address
  if ( verbosity != 0) printf("IDE(0x%x:%d)\n", CF_COMMON_MEM_BASE, interface);

  // Reset the interface
  cfReset();

  // Having found a device, interogate it for its parameters
   if (verbosity != 0)
   {
      if (!ide_ident( interface, (unsigned short *)id_buf))
      {
         return -1;
      }

      id_strcpy((char *)ident.serial,       ide_idData->serial,       20);
      id_strcpy((char *)ident.firmware_rev, ide_idData->firmware_rev,  8);
      id_strcpy((char *)ident.model_num,    ide_idData->model_num,    40);

      ident.cylinders_num   = ide_idData->num_cylinders;
      ident.heads_num       = ide_idData->num_heads;
      ident.sectors_num     = ide_idData->num_sectors;
      ident.lba_sectors_num = ide_idData->lba_total_sectors[1] << 16 | ide_idData->lba_total_sectors[0];

      printf("\tSerial        : %s\n", ident.serial);
      printf("\tFirmware Rev. : %s\n", ident.firmware_rev);
      printf("\tModel         : %s\n", ident.model_num);
      printf("\tC/H/S         : %d/%d/%d\n", ident.cylinders_num, ident.heads_num, ident.sectors_num);
      printf("\tKind          : %x\n", (ide_idData->general_conf>>8)&0x1f);
      printf("\tMode          : %d\n", (ide_idData->pio_xferc_imting >> 8));
   }

  return rval;
}

int  cfRead(int interface, char *buf, int blknum, int blkcount, int verbose)
{
    int                   j;
    unsigned short int    p;
    unsigned char  *      b=(unsigned char *)buf;

    interface = interface;
    
    if ( wait_for_drq())
	{
		return 1;
	}

    CF_REG_WR( IDE_REG_COUNT,    blkcount);    // count =1
    CF_REG_WR( IDE_REG_LBALOW,   blknum        & 0xff);
    CF_REG_WR( IDE_REG_LBAMID,  (blknum >>  8) & 0xff);
    CF_REG_WR( IDE_REG_LBAHI,   (blknum >> 16) & 0xff);
    CF_REG_WR( IDE_REG_DEVICE, ((blknum >> 24) & 0xf) | (0 << 4) | 0xe0);
    CF_CMD_WR( ReadSectorWithRetries );

	monDelay(5);

   if (!wait_for_drq( ))
        return 1;
    //
    // It would be fine if all buffers were word aligned,
    // but who knows
    //
    for (j = 0; j < ( 256 ); j++)
    {
        CF_REG_RD16( IDE_REG_DATA, p);
        
        *b++=p&0xff;
        *b++=(p>>8)&0xff;
    }
    
    return 0;
}

int  cfWrite(int interface, char *buf, int blknum, int blkcount, int verbose)
{
  int                 j;
  unsigned short int  p;
  unsigned char  *    b=(unsigned char *)buf;
  
  unsigned char       status, reason;
  

  if (!wait_for_idle())
  {
    return 1;
  }

  CF_REG_WR( IDE_REG_COUNT, blkcount);    // count =1
  CF_REG_WR( IDE_REG_LBALOW,  blknum & 0xff);
  CF_REG_WR( IDE_REG_LBAMID, (blknum >>  8) & 0xff);
  CF_REG_WR( IDE_REG_LBAHI,  (blknum >> 16) & 0xff);
  CF_REG_WR( IDE_REG_DEVICE,((blknum >> 24) & 0xf) | (0 << 4) | 0x40 | 0xA0);
  CF_REG_WR( IDE_REG_COMMAND, 0x30);

  if (!wait_for_drq( ))
  {
    return 1;
  }

  // It would be fine if all buffers were word aligned, but who knows.

  for (j = 0; j < ( SECTOR_SIZE / sizeof( unsigned short int)); j++)
  {
    p  =  *b++;
    p |= (*b++ << 8);

    CF_REG_WR16( IDE_REG_DATA, p);
  }
  
   // wait for not busy transferring data
    do {
        CF_REG_RD( IDE_REG_STATUS, status);
        CF_REG_RD( IDE_REG_REASON, reason);

        if ((status & (IDE_STAT_BSY | IDE_STAT_DRQ)) == IDE_STAT_DRQ)
        {
          if (reason & IDE_REASON_COD)
          {
            continue;  // still wanting packet data (should timeout here)
          }
        }  
    } while (status & IDE_STAT_BSY);
   

  return 0;
}

