/*    .file "rom_reset.s" */

/*
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 * 
 *
 * Modified for the CSB351 Au1100 Single Board
 * Mike Kelly - Cogent Computer Systems, Inc.
 *
 * rom_reset.s:
  */

#include "config.h"
#include "cpu.h"
#include "mips.h"
#include "au1100.h"
#include "cpuio.h"
#include "warmstart.h"

	.text
	.set noreorder

    .globl  coldstart
    .globl  warmstart
    .globl  moncomptr
    .extern start
	.extern exc080
	.extern exc100
	.extern exc200
	.extern exc280
	.extern exc300
	.extern exc380
	.extern exc400
	.extern exc480

	.section .vectors
    .text
	.global reset

/* coldstart:
 * Entrypoint for the RAM-based monitor image.  
 * If we get here, then we know that we took some kind of "hard" reset.
 * Some register must be preserved from here until the point where we
 * call start so that we can pass the state to start().  Since we got
 * here, we will load the register with INITIALIZE to indicate the fact
 * that we took a hard reset (or powerup).
 * Note that the register used should be the register that the compiler 
 * would normally populate if calling a function with a single argument.
 * In this case, it is t7.
 */
coldstart:
	li    	t7, INITIALIZE
	b		warmstart_more
	nop

/* moncomptr:
 * This tag is at a fixed offset from base of the image so that is easily
 * accessible as a "well-known-address".  This address should be the
 * base of the ram image plus 0x10.
 * The longs after moncomptr (and coldstart above) are there just as a
 * convenient mechanism to look at the image to verify proper placement
 * of the starting point.
 */
	.balign 0x10
#include "moncomptr.S"

/* Different MIPS CPUS use different exception offsets.
 * For now we simply support the general exception and 
 * print status for others...
 * Certainly could be made more robust!
 * Note that the code between startVtable and endVtable could be relocated
 * (by vinit()) if necessary.
 */
	.balign 0x80
	.global startVtable
startVtable:

	la	k0, exc080
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc100
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	j asm_exception
	nop

	.balign 0x80
	la	k0, exc200
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc280
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc300
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc380
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc400
	KSEG1A(k0)
	j	k0
	nop

	.balign 0x80
	la	k0, exc480
	KSEG1A(k0)
	j	k0
	nop

asm_exception:
	addiu	sp, sp, -4
	sw	k0, (sp)
	la	k0, regtbl
	sw	zero, (k0)
	.set noat
	sw	r1, 1*4(k0)
	.set at
	sw	r2, 2*4(k0)
	sw	r3, 3*4(k0)
	sw	r4, 4*4(k0)
	sw	r5, 5*4(k0)
	sw	r6, 6*4(k0)
	sw	r7, 7*4(k0)
	sw	r8, 8*4(k0)
	sw	r9, 9*4(k0)
	sw	r10,10*4(k0)
	sw	r11,11*4(k0)
	sw	r12,12*4(k0)
	sw	r13,13*4(k0)
	sw	r14,14*4(k0)
	sw	r15,15*4(k0)
	sw	r16,16*4(k0)
	sw	r17,17*4(k0)
	sw	r18,18*4(k0)
	sw	r19,19*4(k0)
	sw	r20,20*4(k0)
	sw	r21,21*4(k0)
	sw	r22,22*4(k0)
	sw	r23,23*4(k0)
	sw	r24,24*4(k0)
	sw	r25,25*4(k0)
	sw	r26,26*4(k0)
	sw	r27,27*4(k0)
	sw	r28,28*4(k0)
	sw	r29,29*4(k0)
	sw	r30,30*4(k0)
	sw	r31,31*4(k0)

	/* Load CAUSE register so that it can be passed as an argument
	 * to the exception() C function...
	 */
	mfc0 a0, C0_CAUSE
	la	k0, exception
	KSEG1A(k0)
	j	k0
	nop

	.global endVtable
endVtable:

/* warmstart:
 * This function can be entered either by having previously hit coldstart,
 * or by the monitor source code calling warmstart().  It is assumed to
 * be a function to the C-code, and it is passed a reason for the warmstart.
 * The register which the compiler uses for storage of the calling argument
 * must be saved so that it can be passed to start().
 */
	.align 3
warmstart:
	addiu	t7,a0,0	

warmstart_more:
	/*
	 * Establish Status Register
	 * We're in RAM now, so clear the BEV so that we use
	 * the RAM-based exception handlers...
	 */
	li		t1, 0x00000000
	mtc0	t1, CP0_Status

	/*
	 * Establish CP0 Config0
	 * (set K0=3)
	 */
	li		t1, 0x00000003
	mtc0	t1, CP0_Config0

	/*
	 * Disable Watchpoint facilities
	 * REMOVED FOR MICROCROSS COMPILATION
	li 		t1, 0x00000000
	mtc0	t1, CP0_WatchLo
	mtc0	t1, CP0_IWatchLo
	 */

	/*
	 * Disable the performance counters
	 * REMOVED FOR MICROCROSS COMPILATION
	mtc0	zero, CP0_PerfCtrl
	nop
	 */

	/*
	 * Establish EJTAG Debug register
	 */
	mtc0	zero, CP0_Debug
	nop

	/*
	 * Establish Cause
	 * (set IV bit)
	 */
	li		t1, 0x00800000
	mtc0	t1, CP0_Cause

	/*
	 * Initialize the TLB
	 */
	li		t0, 0		 	# index value
	li		t1, 0x00000000 	# entryhi value
	li		t2, 32		   	# 32 entries
tlbloop:
	/* Probe TLB for matching EntryHi */
	mtc0	t1, CP0_EntryHi
	tlbp
	nop

	/* Examine Index[P], 1=no matching entry */
	mfc0	t3, CP0_Index
	li		t4, 0x80000000
	and		t3, t4, t3
	addiu	t1, t1, 1		# increment t1 (asid)
	beq		zero, t3, tlbloop
	nop

	/* Initialize the TLB entry */
	mtc0	t0, CP0_Index
	mtc0	zero, CP0_EntryLo0
	mtc0	zero, CP0_EntryLo1
	mtc0	zero, CP0_PageMask
	tlbwi

	/* Do it again */
	addiu	t0, t0, 1	
	bne		t0, t2, tlbloop
	nop

	/* Establish Wired (and Random) */
	mtc0	zero, CP0_Wired
	nop

	/* Create a stack using the space designated by monStackEnd
	 */
	la		sp, MonStack
	addiu	sp, MONSTACKSIZE
	addiu 	sp, -4*4

goToC:
	addiu	a0,t7,0
	la		k0, start
	KSEG1A(k0)
	j		k0
	nop
