/*********************************************************************
 *
 * Copyright:
 *	MOTOROLA, INC. All Rights Reserved.  
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Motorola, Inc. This 
 *  software is provided on an "AS IS" basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, MOTOROLA 
 *  DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 *  PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE 
 *  SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY 
 *  ACCOMPANYING WRITTEN MATERIALS.
 * 
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL MOTOROLA BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING 
 *  WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS 
 *  INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY
 *  LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.   
 * 
 *  Motorola assumes no responsibility for the maintenance and support
 *  of this software
 ********************************************************************/

/*
 * File:	nbuf.c
 * Purpose:	Implementation of network buffer scheme.
 *
 * Notes:	These routines implement a static buffer scheme.
 *			The buffer descriptors are as specified by the 
 *			MPC860T/MCF5272 FEC. These routines require that
 *			every packet to be transmitted contains less than
 *			or equal to TX_BUFFER_SIZE bytes.
 *
 *			These routines were written to be used with generic
 *			NE2000 drivers as well as MPC860T/MCF5272 drivers.
 *
 */

#include "config.h"
#if INCLUDE_ETHERNET
#include "nbuf.h"
#include "mx35.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "endian.h"

//#ifdef DONT_USE_SRAM
//#undef DONT_USE_SRAM
//#endif

/********************************************************************/

/* Buffer descriptor indexes */
uint8_t iRxbd;
uint8_t iTxbd;

/* Buffer Descriptors -- must be aligned on a 4-byte boundary but a 
 * 16-byte boundary is recommended. To avoid playing games with the 
 * various compilers and their different extension to ANSI C, these 
 * buffers are aligned by allocating an extra line of data and 
 * adjusting the pointers in nbuf_init().
 *
 * NOTE: according to the documentation...
 *	Buffer descriptor rings must start on a 128-bit boundary.
 * Not sure what that means, but this code assumes it means a
 * 128-byte boundary; hence the use of 256 below (just to be safe).
 */
uint8_t unaligned_txbd[(sizeof(NBUF) * NUM_TXBDS) + 256];
uint8_t unaligned_rxbd[(sizeof(NBUF) * NUM_RXBDS) + 256];

NBUF *TxNBUF;
NBUF *RxNBUF;

/* Data Buffers -- must be aligned on a 16-byte boundary. To avoid 
 * playing games with the various compilers and their different 
 * extension to ANSI C, these buffers are aligned by allocating an 
 * extra line of data and adjusting the pointers in nbuf_init().
 */
uint8_t unaligned_txbuffer[(TX_BUFFER_SIZE * NUM_TXBDS) + 256];
uint8_t unaligned_rxbuffer[(RX_BUFFER_SIZE * NUM_RXBDS) + 256];

uint8_t *TxBuffer;
uint8_t *RxBuffer;

/********************************************************************/
void
nbuf_init ()
{
	uint8_t i;

	/* Clear all buffers at startup...
	 */
	memset((char *)unaligned_txbuffer,0,sizeof(unaligned_txbuffer));
	memset((char *)unaligned_rxbuffer,0,sizeof(unaligned_rxbuffer));
	memset((char *)unaligned_txbd,0,sizeof(unaligned_txbd));
	memset((char *)unaligned_rxbd,0,sizeof(unaligned_rxbd));

#if defined(DONT_USE_SRAM)
	TxNBUF = (NBUF *)((uint32_t)(unaligned_txbd + 256) & 0xFFFFFF00);
	RxNBUF = (NBUF *)((uint32_t)(unaligned_rxbd + 256) & 0xFFFFFF00);
#else
	TxNBUF = (NBUF *) INT_SRAM_ADDRESS;
	RxNBUF = (NBUF *) TxNBUF + NUM_TXBDS;
#endif

	TxBuffer = (uint8_t *)((uint32_t)(unaligned_txbuffer + 256) & 0xFFFFFF00);
	RxBuffer = (uint8_t *)((uint32_t)(unaligned_rxbuffer + 256) & 0xFFFFFF00);

	if (EtherVerbose & SHOW_DRIVER_DEBUG) {
		printf("unaligned_txbuf: 0x%08lx, unaligned_rxbuf: 0x%08lx\n",
			(long)unaligned_txbuffer,(long)unaligned_rxbuffer);
		printf("unaligned_txbd: 0x%08lx, unaligned_rxbd: 0x%08lx\n",
			(long)unaligned_txbd,(long)unaligned_rxbd);
		printf("TxNBUF: 0x%08lx, RxNBUF: 0x%08lx\n",
			(long)TxNBUF,(long)RxNBUF);
		printf("TxBuffer: 0x%08lx, RxBuffer: 0x%08lx\n",
			(long)TxBuffer,(long)RxBuffer);
		printf("sizeof (NBUF) = %d\n",sizeof(NBUF));
	}

	/* Initialize receive descriptor ring */
	for (i = 0; i < NUM_RXBDS; i++)
	{
		RxNBUF[i].status = MCF_FEC_RxBD_E;
		RxNBUF[i].length = 0;
		RxNBUF[i].data = &RxBuffer[i * RX_BUFFER_SIZE];
	}
	/* Set the Wrap bit on the last one in the ring */
	RxNBUF[NUM_RXBDS - 1].status |= MCF_FEC_RxBD_W;

	/* flush cache for the recieve buffers */
	flushDcache((char *) RxBuffer, NUM_RXBDS * RX_BUFFER_SIZE);

	/* Initialize transmit descriptor ring */
	for (i = 0; i < NUM_TXBDS; i++)
	{
		TxNBUF[i].status = MCF_FEC_TxBD_L | MCF_FEC_TxBD_TC;
		TxNBUF[i].length = 0;
		TxNBUF[i].data = &TxBuffer[i * TX_BUFFER_SIZE];
	}
	/* Set the Wrap bit on the last one in the ring */
	TxNBUF[NUM_TXBDS - 1].status |= MCF_FEC_TxBD_W;

	/* Initialize the buffer descriptor indexes */
	iTxbd = iRxbd = 0;

	return;
}
/********************************************************************/
uint32_t
nbuf_get_start(uint8_t direction)
{
	/* 
	 * Return the address of the first buffer descriptor in the ring.
	 * This routine is needed by the FEC of the MPC860T and MCF5272
	 * in order to write the Rx/Tx descriptor ring start registers
	 */
	switch (direction){
	case Rx:
		return (uint32_t)RxNBUF;
	case Tx:
	default:
		return (uint32_t)TxNBUF;
	}
}

/********************************************************************/
NBUF *
nbuf_rx_allocate ()
{
	/* Return a pointer to the next empty Rx Buffer Descriptor */
	int i = iRxbd;

	/* Check to see if the ring of BDs is full */
	if (RxNBUF[i].status & MCF_FEC_RxBD_INUSE)
		return 0;

	/* Mark the buffer as in use */
	RxNBUF[i].status |= MCF_FEC_RxBD_INUSE;
	
	/* increment the circular index */
	iRxbd = (uint8_t)((iRxbd + 1) % NUM_RXBDS);

	return &RxNBUF[i];
}

/********************************************************************/
NBUF *
nbuf_tx_allocate ()
{
	/* Return a pointer to the next empty Tx Buffer Descriptor */
	int i = iTxbd;

	/* Check to see if ring of BDs is full */
	if (TxNBUF[i].status & MCF_FEC_TxBD_BUSY)
		return 0;
	
	/* Mark the buffer as Ready (in use) */
	/* FEC must set R bit in transmit routine */
	TxNBUF[i].status |= MCF_FEC_TxBD_BUSY;

	/* increment the circular index */
	iTxbd = (uint8_t)((iTxbd + 1) % NUM_TXBDS);

	return &TxNBUF[i];
}

/********************************************************************/
void
nbuf_rx_release (NBUF* pNbuf)
{
	/* Mark the buffer as empty and not in use */
	pNbuf->status |= MCF_FEC_RxBD_E;
	pNbuf->status &= ~MCF_FEC_RxBD_INUSE;

	flushDcache((char *) pNbuf->data, pNbuf->length);
}

/********************************************************************/
void
nbuf_tx_release (NBUF* pNbuf)
{
	/* Mark the buffer as not in use */
	pNbuf->status &= ~MCF_FEC_TxBD_BUSY;

	flushDcache((char *) pNbuf->data, pNbuf->length);
}

/********************************************************************/
int
nbuf_rx_next_ready()
{
	/****************************************************************
	 This function checks the EMPTY bit of the next Rx buffer to be
	 allocated. If the EMPTY bit is cleared, then the next buffer in
	 the ring has been filled by the FEC and has not already been
	 allocated and passed up the stack. In this case, the next buffer
	 in the ring is ready to be allocated. Otherwise, the  buffer is
	 either empty or not empty but still in use by a higher level
	 protocol. The FEC receive routine uses this function to determine
	 if multiple buffers where filled by the FEC during a single
	 interrupt event.
	 ****************************************************************/

	return ( !(RxNBUF[iRxbd].status & MCF_FEC_RxBD_E));
}

int
nbuf_rx_overflow()
{
	return (RxNBUF[iRxbd].status & MCF_FEC_RxBD_OV);
}

int
nbuf_rx_lastinframe()
{
	return (RxNBUF[iRxbd].status & MCF_FEC_RxBD_L);
}

int
nbuf_rx_status()
{
	int i;

	for (i = 0; i < NUM_RXBDS; i++)
		printf("0x%x\n",RxNBUF[i].status);
	return (RxNBUF[iRxbd].status);
}

#endif

