//==========================================================================
//
// cpu_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains code to intialize the PXA250 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//
// Modified by Luis Torrico, Cogent Computer Systems, Inc.
//	03.28.2007 : Added all GPIO Settings for CSB726 PXA270 Cogent board.
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "cpu_gpio.h"
#include "mpc5121e.h"
#include "umongpio.h"

// Function Prototypes
void GPIO_init(void);
int GPIO_set(int);
int GPIO_clr(int);
int GPIO_tst(int);
int GPIO_in(int);
int GPIO_out(int);

//--------------------------------------------------------------------------
//
// GPIO assignments for the CSB726 are as follows.
// Active indicates what level is considered active for that
// signal, if any.  A/F indicates alternate function mode (0 = none).
//
// GPIO		Direction	Active	A/F	Description
// 0		Input		Low		0	*PWR_INT - Power Interrupt from AVR
// 1		I/O			Low		0	GPIO 0 - used as *IRQ to the processor
// 2		Unused
// 3		I/O			N/A		0	PWR_SCL - Power Manager I2C SCL for ISL6271 Programmable Core Regulator
// 4		I/O			N/A		0	PWR_SDA - Power Manager I2C SDA for ISL6271 Programmable Core Regulator
// 5-8		Unused
// 9		I/O			N/A		0	GPIO 2 - Can be used as Timer Channel Output 0
// 10		I/O			N/A		0	GPIO 3 - Can be used as Timer Channel Output 1
// 11		I/O			Low		0	GPIO 1 - Can be used as *PME
// 12		Input		Low		0	*PIRQ - Touch Controller Interrupt
// 13		Output		N/A		1	SPI0_MOSI - Synchronous Serial Port 0 Transmit
// 14		Output		Low		2	*SPI0_CS0 - Synchronous Serial Port 0 Frame
// 15		Input		Low		0	*EXP_IRQ - Used as Expansion Interrupt from the CPLD
// 16		I/O			N/A		0	GPIO 4 - Can be used as PWM0
// 17		I/O			N/A		0	GPIO 5 - Can be used as PWM1
// 18		Input		Low		1	*EXP_WAIT, External Bus Ready (PXA270 RDY)
// 19		Output		N/A		1	SPI0_CLK - Synchronous Serial Port 0 Clock
// 20		Unused					This GPIO is tied to GPIO 109 so it is unavailable.
// 21		Unused		
// 22		Input		N/A		1	SSP2_EXTCLK - Synchronous Serial Port External Clock for SPI0
// 23-27	Unused					Reserved for Quick Capture Interface
// 28		Input		N/A		1	AC_BCLK - AC97 Bit Clock from Codec
// 29		Input		N/A		1	AC_SDIN - AC97 Serial Data from Codec
// 30		Output		N/A		2	AC_SDOUT - AC97 Serial Data to Codec
// 31		Output		N/A		2	AC_SYNC - AC97 SYNC to Codec
// 32		Output		N/A		2	SD_CLK - MMC Clock
// 33		Output		Low		2	Chip Select 5 - Unused
// 34		Input		N/A		1	U0_RXD - Full Function UART RXD
// 35		Input		N/A		1	U0_CTS - Full Function UART CTS
// 36		Input		N/A		1	U0_DCD - Full Function UART DCD
// 37		Input		N/A		1	U0_DSR - Full Function UART DSR
// 38		Input		N/A		1	U0_RI - Full Function UART RI
// 39		Output		N/A		2	U0_TXD - Full Function UART TXD
// 40		Output		N/A		2	U0_DTR - Full Function UART DTR
// 41		Output		N/A		2	U0_RTS - Full Function UART RTS
// 42		Input		N/A		1	U1_RXD - Bluetooth UART RXD
// 43		Output		N/A		2	U1_TXD - Bluetooth UART TXD
// 44		Input		N/A		1	U1_CTS - Bluetooth UART CTS
// 45		Output		N/A		2	U1_RTS - Bluetooth UART RTS
// 46		Input		N/A		2	D_RXD (Debug UART) - Standard UART RXD
// 47		Output		N/A		1	D_TXD (Debug UART) - Standard UART TXD
// 48		Output		Low		2	*CF_OE - Compact FLASH Output Enable
// 49		Output		Low		2	*CF_WE - Compact FLASH Write Enable
// 50		Output		Low		2	*CF_IOR - Compact FLASH I/O Read Enable	 
// 51		Output		Low		2	*CF_IOW - Compact FLASH I/O Write Enable
// 52		Input		Low		0	*E_INT - LAN9215 Interrupt
// 53		Input		Low		0	*SM_INT - SM501 Interrupt
// 54		Output		Low		2	*CF_CE2 - Compact Flash Chip Enable 2
// 55		Output		Low		2	*CF_REG - Compact Flash Register Select
// 56		Input		Low		1	*CF_WAIT - Compact Flash Wait Request
// 57		Unused
// 58-77	Unused					Reserved for LCD Interface
// 78		Output		Low		2	*EXP_CS - Chip Select 2 - Expansion CS to CPLD
// 79		Output		Low		2	Chip Select 3 - LAN9215
// 80		Output		Low		2	Chip Select 4 - SM501
// 81		Output		N/A		1	SPI1_MOSI - Synchronous Serial Port 1 Transmit
// 82		Input		N/A		1	SPI1_MOSI - Synchronous Serial Port 1 Receive
// 83		Output		Low		1	*SPI1_CS0 - Synchronous Serial Port 1 Frame
// 84		Output		N/A		1	SPI1_CLK - Synchronous Serial Port 1 Clock
// 85		Output		Low		1	*CF_CE1 - Compact Flash Chip Enable 1
// 86		Input		Low		1	SPI0_MOSI - Synchronous Serial Port 0 Receive
// 87		Output		High	0	PWR_1WIRE
// 88		Unused
// 89		Unused
// 90		Unused					Reserved for Quick Capture Interface
// 91		Unused					Reserved for Quick Capture Interface
// 92		I/O			N/A		1	SD_D0 - MMC Data Bit 0
// 93		Unused					Reserved for Quick Capture Interface
// 94		I/O 		N/A		0	GPIO 6
// 95		I/O 		N/A		0	GPIO 7
// 96		I/O 		N/A		0	GPIO 8
// 97		I/O 		N/A		0	GPIO 9
// 98		Input		High	0	CF_RDY - Compact FLASH Card Ready/Busy and Interrupt
// 99		Input		Low		0	*CF_CD - Compact FLASH Card Detect
// 100		Input		Low		0	*SD_CD - MMC Card Detect
// 101		Input		Low		0	*SD_WP - MMC Write Protect
// 102		Unused
// 103		Output		High	0	CF_RST - Compact Flash Card Reset
// 104		Output		Low		0	PWR_OFF - Power Off request to AVR
// 105		Unused
// 106		Unused
// 107		Unused
// 108		Unused					Reserved for Quick Capture Interface
// 109		I/O 		N/A		1	SD_D1 - MMC Data Bit 1 (MMC Interrupt)
// 110		I/O 		N/A		1	SD_D2 - MMC Data Bit 2
// 111		I/O 		N/A		1	SD_D3 - MMC Data Bit 3
// 112		Output		N/A		1	SD_CMD - MMC Command
// 113		Output		Low		2	*AC_RST - AC97 Reset to Codec
// 114		Unused					Reserved for Quick Capture Interface
// 115		Unused					Reserved for Quick Capture Interface
// 116		Unused					Reserved for Quick Capture Interface
// 117		I/O 		N/A		1	I2C_SCL - I2C Clock
// 118		I/O 		N/A		1	I2C_SDA - I2C Data

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the MPC5121e
// as used on the CSB781
//
void
GPIO_init()
{

	// 
	GPIO_REG(GPIO_OD) = 0x41000000;
	GPIO_REG(GPIO_IER) = 0xffffffff;
	GPIO_REG(GPIO_IMR) = 0xffffffff;

	// Next, set the correct direction for our functions.  A one means output.
	//GPIO_REG(GPIO_DIR) = 0xffffffff;	   
//GPIO_out(GPIO_7);
//GPIO_out(GPIO_1);

//GPIO_in(GPIO_7);
//GPIO_in(GPIO_1);

//GPIO_out(GPIO_7);
//GPIO_out(GPIO_1);

//GPIO_in(GPIO_7);
//GPIO_in(GPIO_1);
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if setting the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_set(int gpio_bit)
{
	// Get the direction and see if it's an output
//	if (!(GPIO_REG(GPIO_DIR) & (1 << gpio_bit)))
//		return -1;

	// if yes, then set it
	GPIO_REG(GPIO_DATA) |= (1 << gpio_bit);
	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if clearing the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_clr(int gpio_bit)
{
	// Get the direction and see if it's an output
//	if (!(GPIO_REG(GPIO_DIR) & (1 << gpio_bit)))
//		return -1;

	// if yes, then set it
	GPIO_REG(GPIO_DATA) &= ~(1 << gpio_bit);
	return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
	// Get the state
	if (GPIO_REG(GPIO_DATA) & (1 << gpio_bit)) return 1;
	
	return 0; // bit was not set
}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_in(int gpio_bit)
{
	//GPIO_REG(GPIO_DATA) &= ~(1 << gpio_bit);
	GPIO_REG(GPIO_DIR) &= ~(1 << gpio_bit);
	return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_out(int gpio_bit)
{
	//GPIO_REG(GPIO_DATA) &= ~(1 << gpio_bit);
	GPIO_REG(GPIO_DIR) |= (1 << gpio_bit);
	return 0;
}

