/* cpuio.c:

	General notice:
	This code is part of a boot-monitor package developed as a generic base
	platform for embedded system designs.  As such, it is likely to be
	distributed to various projects beyond the control of the original
	author.  Please notify the author of any enhancements made or bugs found
	so that all may benefit from the changes.  In addition, notification back
	to the author will allow the new user to pick up changes that may have
	been made by other users after this version of the code was distributed.

	Author:	Ed Sutter
	email:	esutter@lucent.com		(home: lesutter@worldnet.att.net)
	phone:	908-582-2351			(home: 908-889-5161)
*/
#include "config.h"
#include "cpu.h"
#include "cpuio.h"
#include "ether.h"
#include "genlib.h"
#include "stddefs.h"
#include "warmstart.h"

struct	bd *rxbd;
uchar	xbuf[8], rbuf[8];
int	ConsoleBaudRate;

void CSSetup(void);
extern int getreg(), putreg();
extern int putsr(), getsr();

int
ConsoleBaudSet(int baud)
{
	switch(baud) {			/* Baud rate options with 20Mhz xtal */
		case 9600:			/* CD = (20000000/(9600*16)) - 1 */
			*(ushort *)SCON1 = (0|(129<<1));
			break;
		case 38400:			/* CD = (20000000/(38400*16)) - 1 */
			*(ushort *)SCON1 = (0|(32<<1));
			break;
		case 19200:			/* CD = (20000000/(19200*16)) - 1 */
		default:
			*(ushort *)SCON1 = (0|(64<<1));
			break;
	}
	return(0);
}

/* InitUART():
	This function is called at startup, application exit, or by the 
	mstat -b BAUD command.
	For startup and mstat, the baud rate will be specified and used; for
	application exit, it will be 0 to indicate that whatever baud rate was
	used prior by mstat or startup should be used again.
*/
void
InitUART(int baud)
{

	if (baud == 0)
		baud = ConsoleBaudRate;

	/* Keep track of the last baud rate, so that it can be used if the */
	/* incoming baudrate is NULL. */
	ConsoleBaudRate = baud;

	/* Set up SCC1... */
	/* SIMODE: See page 4-23 */
	*(ushort *)SIMODE = 0x0000;	/* Non-muxed Serial Interface Mode */

	ConsoleBaudSet(baud);
	
	/* SCM: See pages 4-31 and 4-64. */
	*(ushort *)SCM1 = 0x01b1;	/* Software CTS&CD, Async mode. */
	/* No parity, 1 stop bit */

	*(uchar *)RFCR1 = 0;		/* Transmit & receive function codes. */
	*(uchar *)TFCR1 = 0;

	/* RCCR: page 4-59 */
	*(ushort *)RCCR1C1 = 0x8000;	/* Clear UART Control char regs...*/
	*(ushort *)RCCR1C8 = 0x0000;

	*(ushort *)PAREC1 = 0;			/* Clear error counts */
	*(ushort *)FRMEC1 = 0;
	*(ushort *)NOSEC1 = 0;
	*(ushort *)BRKEC1 = 0;

	*(ushort *)UADR11 = 0;			/* Clear UART address (not used) */
	*(ushort *)UADR21 = 0;

	*(ushort *)MRBLR1 = 1;			/* Maximum rcv buffer length */
	*(uchar *)RBDNM1 = 0x00;		/* Rcv buffer descriptor number */
	*(uchar *)TBDNM1 = 0x40;		/* Xmt buffer descriptor number */

	/* UART xmtbuf descriptors: pg 4-71 */
	*(ushort *)TXBD0_0 = 0x6000;	/* Status and control */
	*(ushort *)TXBD0_1 = 0x0001;	/* Data length */
	*(uchar **)TXBD0_2 = &xbuf[0];	/* Buffer pointer */

	/* UART rcvbuf descriptors: pg 4-66 */
	/* Set up a 4-byte fifo. */
#if 0	/* Uses external memory for data: */
	*(ushort *)RXBD0_0 = 0xd000;	/* Status and control */
	*(ushort *)RXBD0_1 = 0x0001;	/* Data length */
	*(uchar **)RXBD0_2 = &rbuf[0];	/* Buffer pointer */
	*(ushort *)RXBD1_0 = 0xd000;	/* Repeat for 3 more BDs... */
	*(ushort *)RXBD1_1 = 0x0001;
	*(uchar **)RXBD1_2 = &rbuf[1];
	*(ushort *)RXBD2_0 = 0xd000;
	*(ushort *)RXBD2_1 = 0x0001;
	*(uchar **)RXBD2_2 = &rbuf[2];
	*(ushort *)RXBD3_0 = 0xf000;
	*(ushort *)RXBD3_1 = 0x0001;
	*(uchar **)RXBD3_2 = &rbuf[3];
#else 	/* Uses internal memory for data: */
	*(ushort *)RXBD0_0 = 0x9000;	/* Status and control */
	*(ushort *)RXBD0_1 = 0x0001;	/* Data length */
	*(uchar **)RXBD0_2 = (uchar *)0;/* Buffer pointer */
	*(ushort *)RXBD1_0 = 0x9000;	/* Repeat for 3 more BDs... */
	*(ushort *)RXBD1_1 = 0x0001;
	*(uchar **)RXBD1_2 = (uchar *)1;
	*(ushort *)RXBD2_0 = 0x9000;
	*(ushort *)RXBD2_1 = 0x0001;
	*(uchar **)RXBD2_2 = (uchar *)2;
	*(ushort *)RXBD3_0 = 0xb000;
	*(ushort *)RXBD3_1 = 0x0001;
	*(uchar **)RXBD3_2 = (uchar *)3;
#endif
	rxbd = (struct bd *)RXBD0_0;

	/* SCCM: page 4-76 */
	*(uchar *)SCCM1 = 0xff;				/* Enable all events */

	/* SCCE: page 4-73 */
	*(uchar *)SCCE1 = 0xff;				/* Clear all events */

	*(ushort *)SCM1 = 0x01b1 | 0x000c;	/* Enable transmit and receive */

	*(uchar *)SCCE1 = *(uchar *)SCCE1;

	InitRemoteIO();
}


/* The common version of start.c calls devInit(baud) instead of the older 
 * InitUART(baud).  To become compatible with this, we encapsulate InitUART()
 * with devInit() here... 
 */
int
devInit(int baud)
{
	InitUART(baud);
	return(0);
}

void
initCPUio()
{
	*(ushort *)BAR = 0x0700;	/* reinit BAR */
	*(ushort *)MOBAR = 0x0710;	/* reinit MOBAR */
	CSSetup();
}

struct cs {
	ushort	*br;
	ushort	*or;
} cstbl[] = {
		{ (ushort *)BR0, (ushort *)OR0 },
		{ (ushort *)BR1, (ushort *)OR1 },
		{ (ushort *)BR2, (ushort *)OR2 },
		{ (ushort *)BR3, (ushort *)OR3 }
};

char *
rw(br,or)
ushort or,br;
{
	if (or & 0x0002)
		return((br & 0x0002) ? "write-only" : "read-only");
	else
		return("read/write");
}

void
CSInfo()
{
	int	csnum, i;
	ushort	br, or;

	printf("\nCurrent Chip Select Ctrl-Regs:\n");
	printf(" BR0: 0x%04x,  BR1: 0x%04x,  BR2: 0x%04x,  BR3: 0x%04x\n",
	    *(ushort *)BR0,*(ushort *)BR1,*(ushort *)BR2,*(ushort *)BR3);
	printf(" OR0: 0x%04x,  OR1: 0x%04x,  OR2: 0x%04x,  OR3: 0x%04x\n",
	    *(ushort *)OR0,*(ushort *)OR1,*(ushort *)OR2,*(ushort *)OR3);

	printf("\nCurrent Chip Select Config:\n");
	for(i=0;i<4;i++) {
		csnum = i;
		br = *(cstbl[i].br);
		or = *(cstbl[i].or);
		printf(" CS%d base: 0x%06lx %s, Mask: 0x%04lx, FC: %d%d%d%s, %s\n",
		    i, (ulong)(br & 0x1ffc) << 11,
		    (br & 0x0001) ? " (enabled)" : "(disabled)",
		    (ulong)(or & 0x1ffc) << 11,
		    (br & 0x8000) ? 1 : 0,
		    (br & 0x4000) ? 1 : 0,
		    (br & 0x2000) ? 1 : 0,
		    (or & 0x0001) ? "" : " (ignored)",
		    rw(br,or));
		/* Assumes 68000 bus cycle description (pg 5-32). */
		if ((or & 0xe000) == 0xe000)
			printf("   (external DTACK, ");
		else
			printf("   (%d wait states",(or & 0xe000) >> 13);
		/*		printf(", %d bit interface",
					*(ushort *)SCRHI & (0x0008 >> i) ? 16 : 8);
		*/
		printf(")\n");
	}
}

void
CSSetup()
{
	/* Chip selects 0 and 1 are already programmed (reset.s).  */
	/* CS0: FLASH 1Mbyte (512Kx16) start @ 0x400000 */
	/* CS1: SRAM 128K (64Kx16) start @ 0x000000 */
	/* CS2: SRAM 128K (64Kx16) start @ 0x020000 */
	/* CS3: DualPortRam 4K (2Kx16) start # 0x800000 */
	*(ushort *)BR3 = 0x1001;
	*(ushort *)OR3 = 0xdf00;
}

int
target_putchar(char c)
{
	/* Wait for ready bit to clear... */
	while(1) {
		if (!(*(ushort *)TXBD0_0 & 0x8000))
			break;
	}
	xbuf[0] = (uchar)c;
	*(ushort *)TXBD0_0 |= 0x8000;
	return((int)c);
}

int	
target_getchar(void)
{
	extern	struct bd *rxbd;
	char	c;

	/* Retrieve character and clear status bit: */
	if (rxbd->stat & 0x4000)	/* If internal, then add BASE. */
		c = *(rxbd->ptr);
	else
		c = *(rxbd->ptr+BASE);

	rxbd->stat |= 0x8000;

	/* If WRAP bit is set in descriptor, reset pointer; else increment: */
	if (rxbd->stat & 0x2000)
		rxbd = (struct bd *)RXBD0_0;
	else
		rxbd++;

	return((int)c);
}

int
target_console_empty(void)
{
	return(1);
}

int
target_gotachar(void)
{
	if (rxbd->stat & 0x8000)
		return(0);
	else
		return(1);
}

void
intsrestore(ulong sr)
{
	putsr(sr);
}

ulong
intsoff(void)
{
	ushort  sr;

	sr = getsr();
	putsr(sr|0x0700);
	return((ulong)sr);
}

void
cacheInitForTarget(void)
{
}

void
target_reset(void)
{
	monrestart(INITIALIZE);
}
