/* ad7490.c:
 * This code supports the BF537's SPI interface connection to 
 * the Analog Devices AD7490 16-channel A-to-D converter (ADC).
 */
#include "config.h"
#include "cdefBF537.h"
#include "genlib.h"
#include "ctype.h"
#include "cli.h"
#include "cpu.h"
#include "spi_bf537.h"
#include "ad7490.h"
#include "ad5384.h"
#include "monlib.h"
#include "stddefs.h"

extern double strtod(char *,char **);
extern char * sprnfloat(char *buf,int bsize,float fpval,int fraclen);

static int ad7490_verbose;
static int ad7490_initialized;
static long ad7490_baud;
unsigned short vinTbl[AD7490_DEVTOT][16];
static unsigned short maxTbl[AD7490_DEVTOT][16];

struct vinInfo {
	char *name;
	char assigned;
};

#if PLATFORM_ROSA
static struct vinInfo vinInfoTbl[] = {
	{ "APD_MON9",	1, },
	{ "APD_MON8",	1, },
	{ "APD_MON7",	1, },
	{ "APD_MON6",	1, },
	{ "APD_MON5",	1, },
	{ "APD_MON4",	1, },
	{ "APD_MON3",	1, },
	{ "APD_MON2",	1, },
	{ "APD_MON1",	1, },
	{ "APD_MON0",	1, },
	{ "RTH",		1, },
	{ "ITEC_MON",	1, },
	{ "AWG_TEMP",	1, },
	{ "NA", 		0, },
	{ "NA", 		0, },
	{ "NA", 		0, }
};

#else 	// PLATFORM_TOSA

static struct vinInfo vinInfoTbl[] = {
	{ "PHOTO_MON9",			1 },
	{ "PHOTO_MON8",			1 },
	{ "PHOTO_MON7",			1 },
	{ "PHOTO_MON6",			1 },
	{ "PHOTO_MON5",			1 },
	{ "PHOTO_MON4",			1 },
	{ "PHOTO_MON3",			1 },
	{ "PHOTO_MON2",			1 },
	{ "PHOTO_MON1",			1 },
	{ "PHOTO_MON0",			1 },
	{ "LOCKER_ITEC_MON",	1 },
	{ "LOCKER_TEMP_MON",	1 },
	{ "TOSA_ITEC_MON",		1 },
	{ "TOSA_TEMP_MON",		1 },
	{ "NA",					0 },
	{ "NA",					0 }
};
#endif

#define AD7490_CTRL	(AD7490_WRITE | RANGE | CODING | PM1 | PM0)

#define MAXVOLTS_FLOAT	((float)2.5)
#define MAXVOLTS_HEX 	0xfff		// AD7490 is 12-bit

int
ad7490vin_to_ad5384vout(int vin)
{
	switch(vin) {
#if PLATFORM_ROSA
	case 9:	
		return(1);
	case 7:	
		return(3);
	case 5:
		return(5);
	case 3:
		return(7);
	case 1:
		return(9);
	case 8:
		return(11);
	case 6:
		return(13);
	case 4:
		return(15);
	case 2:
		return(17);
	case 0:
		return(19);
#else
#endif
	default:
		printf("AD7490vin-to-AD5384vout translation failure (%d)\n",vin);
		return(-1);
	}
}
int
ad7490_voltsTohex(float volts, unsigned short *hex)
{
	if ((volts > MAXVOLTS_FLOAT) || (volts < 0)) {
		printf("Voltage out of range\n");
		return(-1);
	}
	
	*hex = (unsigned short)((volts * MAXVOLTS_HEX)/MAXVOLTS_FLOAT);
	return(0);
}

int
ad7490_hexTovolts(unsigned short hex, float *volts)
{
	if (hex > MAXVOLTS_HEX) {
		printf("Voltage out of range\n");
		return(-1);
	}

	*volts = (float)((float)(MAXVOLTS_FLOAT * (float)hex)/(float)MAXVOLTS_HEX);
	return(0);
}

int
ad7490_write(int dev, unsigned short val)
{
	volatile unsigned short tmp;

	if (ad7490_verbose > 2)
		printf("ad7490_write(%04x)...\n",val);

	SPI_ENABLE();
	SPI_WRITE_MODE();
	*pSPI_FLG &= FLG3;		// Activate chip-select to AD7490 */

	*pSPI_TDBR = val;
	tmp = *pSPI_RDBR;		// Dummy read initiates the WRITE
	SPI_FINWAIT();			// Wait for SPI finish

	*pSPI_FLG |= ~FLG3;		// Deactivate chip-select to AD7490 */
	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
ad7490_read(int dev, unsigned short *val)
{
	volatile unsigned short tmp;

	if (ad7490_verbose > 2)
		printf("ad7490_read()...\n");

	SPI_ENABLE();
	SPI_READ_MODE();

	/* Activate chip-select to AD7490 */
	*pSPI_FLG &= FLG3;

	/* Read 16-bits of data...
	 */
	tmp = *pSPI_RDBR;

	SPI_FINWAIT();

	*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
	SPI_RXSWAIT();			/* Wait for receive buffer full */
	*val = *pSPI_RDBR;

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG3;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

void
ad7490_qinit(int dev)
{
	SPI_ENABLE();

	*pSPI_BAUD = (SYSCLKFREQ/(ad7490_baud*2));
	__builtin_bfin_ssync();

	/* Enable SPI interface as...
	 *  Master, MSB first, 
	 */
	*pSPI_CTL = (SPE | MSTR | CPHA | SIZE);
	__builtin_bfin_ssync();

	SPI_DISABLE();
}

/* ad7490_init():
 * The AD7490 uses SPI_SEL3 on the BF537.
 * It is assumed here that the pin configuration has already been set
 * up during basic board init.
 */
int
ad7490_init(int dev)
{
	int i;
	char *env;

	if (ad7490_verbose > 1)
		printf("ad7490_init()...\n");

	SPI_ENABLE();

	/* The AD7490 is on SPI CS3 line, so enable that line.
	 */
	*pSPI_FLG |= FLS3;
	__builtin_bfin_ssync();

	if ((env = getenv("ADCBAUD")) != (char *)0)
		ad7490_baud = strtol(env,0,0);
	else
		ad7490_baud = ADCBAUD;

	ad7490_qinit(dev);

	*pSPI_STAT = (MODF | TXE | RBSY | TXCOL);
	__builtin_bfin_ssync();

	/* Perform two "dummy" cycles, as described in the datasheet,
	 * then program the control register...
	 */
	ad7490_write(dev, 0xffff);
	ad7490_write(dev, 0xffff);

	ad7490_write(dev, AD7490_CTRL);

	for(i=0;i<16;i++)
		maxTbl[dev][i] = MAXVOLTS_HEX;

	ad7490_initialized = 1;

	SPI_DISABLE();
	return(0);
}

int
getAD7490Chan(char *string, int *chan)
{
	int tmp;

	tmp = atoi(string);
	if ((tmp > 15) || (tmp < 0)) {
		printf("Valid channel range: 0-15\n");
		return(-1);
	}
	if (vinInfoTbl[tmp].assigned == 0)
		printf("Warning: channel not assigned.\n");

	*chan = tmp;
	return(0);
}

int
getBits12(char *string, unsigned short *val)
{
	unsigned short tmp;

	tmp = strtol(string,0,0);
	if (tmp > 0x0fff) {
		printf("value out of 12-bit range\n");
		return(-1);
	}
	*val = tmp;
	return(0);
}


int
ad7490_scan(int dev)
{
	int i, tot = 0;
	unsigned short val;

	for(i=0;i<16;i++) {
		if (vinInfoTbl[i].assigned == 0)
			continue;

		ad7490_write(dev,AD7490_CTRL | ((i << 10) & 0x3c00));
		if (ad7490_read(dev,&val) < 0)
			return(-1);

		vinTbl[dev][(val & 0xf000)>>12] = val & 0xfff;

		if ((val & 0xfff) > maxTbl[dev][i]) {
			ad5384_decrement_vout(dev,ad7490vin_to_ad5384vout(i));
			tot++;
		}
	}
	return(tot);
}

/* overcurrent_scan():
 * This function is called through gotachar() so it is VERY active.
 * Its filtered down a bit by only actually doing something every
 * SCAN calls...
 */
#define SCAN 2500

static int scantot;

void
overcurrent_scan(int dev)
{
	++scantot;
	if (scantot > SCAN) {
		if (ad7490_scan(dev) == 0)
			scantot = 0;
		else
			scantot -= (SCAN/10);
	}
}

char *ad7490Help[] = {
	"Access the AD7490 analog-to-digital converter.",
	"[-d:v] {cmd} [cmd-specific args]",
	" Options:",
	" -d#        specify device number (default=0)",
	" -v         verbosity level (default=0)",
	"",
	" Commands:",
	" init       initialze",
	" vin        read all 16 Vin values",
	" vin {CC}   read Vin value only for channel CC",
	" max {CC} {VV}   set max for channel CC",
	"            (if CC == all, then apply VV to all channels)",
	" scan       scan channels to detect overflow (relative to max setting)",
	0,
};

int
ad7490Cmd(int argc,char *argv[])
{
	char *cmd;
	int devno, opt, rc = 0;

	devno = 0;
	ad7490_verbose = 0;
	while((opt=getopt(argc,argv,"d:v")) != -1) {
		switch(opt) {
		case 'd':
			devno = atoi(optarg);
			break;
		case 'v':
			ad7490_verbose++;
			break;
		default:
			return(CMD_PARAM_ERROR);
		}
	}

	if (devno >= AD7490_DEVTOT) {
		printf("Device num '%d' too big\n",devno);
		return(CMD_FAILURE);
	}
	
	if (argc <= optind)
		return(CMD_PARAM_ERROR);

	cmd = argv[optind];

	if ((argc == optind+1) && (strcmp(cmd,"init") == 0)) {
		ad7490_init(devno);
		return(CMD_SUCCESS);
	}

	if (ad7490_initialized == 0) {
		printf("Run 'ad7490 init' first\n");
		return(CMD_FAILURE);
	}

	ad7490_qinit(devno);

	// The ctrl register can't be read, so if verbosity is high,
	// print out the #define here...
	if (ad7490_verbose > 1)
		printf("CTRL: 0x%04x\n",AD7490_CTRL);

	if (strcmp(cmd,"vin") == 0) {
		float fpval;
		char buf[32];
		char varname[16];

		if (argc == optind+1) {
			int i;

			if (ad7490_scan(devno) == -1) {
				printf("Scan failed\n");
				return(CMD_FAILURE);
			}

			for(i=0;i<16;i++) {
				sprintf(varname,"VIN_%02d",i);
				ad7490_hexTovolts(vinTbl[devno][i],&fpval);
				if (vinInfoTbl[i].assigned) {
					shell_sprintf(varname,"%s",
						sprnfloat(buf,sizeof(buf),fpval,3));
					if (ad7490_verbose) {
						printf("%10s: %sv (hex: 0x%04x)\n",
							vinInfoTbl[i].name,buf,vinTbl[devno][i]);
					}
				}
				else {
					setenv(varname,"NA");
				}
			}
		}
		else if (argc == optind+2) {
			int chan, tot = 0;
			char *range, *env, varname[16];
			unsigned short val;

			range = argv[optind+1];
			for(chan=0;chan<16;chan++) {
				if (!inRange(range,chan))
					continue;
				if (tot++ == 0)
					printf("VIN [%s]: ",range);
				// If varname exists, then print the fake value...
				sprintf(varname,"VIN%d",chan);
				if ((env = getenv(varname)) != (char *)0) {
					printf("%d=%s,",chan,env);
				}
				else {
					ad7490_write(devno, AD7490_CTRL | ((chan << 10) & 0x3c00));
					if ((rc = ad7490_read(devno, &val)) >= 0) {
						ad7490_hexTovolts(val&0xfff,&fpval);
						printf("%d=%s,",
							chan,sprnfloat(buf,sizeof(buf),fpval,3));
				}
				}
			}
			if (tot > 0)
				putchar('\n');
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"max") == 0) {
		int i, chan;
		float fpval;
		char buf[16], *range;
		unsigned short val;

		// If no args after 'max', just display all values:
		if (argc == optind+1) {
			for(i=0;i<16;i++) {
				ad7490_hexTovolts(maxTbl[devno][i],&fpval);
				printf("Chan_%d: %s\n",i,sprnfloat(buf,sizeof(buf),fpval,3));
			}
		}
		// If one arg after 'max', just show that value:
		else if (argc == optind+2) {
			int tot = 0;
			range = argv[optind+1];
			for(chan=0;chan<16;chan++) {
				if (!inRange(range,chan))
					continue;
				if (tot++ == 0)
					printf("MAX [%s]: ",range);
				ad7490_hexTovolts(maxTbl[devno][chan],&fpval);
				printf("%d=%s,",chan,sprnfloat(buf,sizeof(buf),fpval,3));
			}
			if (tot > 0)
				putchar('\n');
		}
		// If one arg after 'max', just set that value:
		else if (argc == optind+3) {
			float fpval;

			range = argv[optind+1];
			fpval = (float)strtod(argv[optind+2],0);
			if (ad7490_voltsTohex(fpval,&val) == -1)
				return(CMD_PARAM_ERROR);

			for (chan=0;chan<16;chan++) {
				if (!inRange(range,chan))
					continue;
				maxTbl[devno][chan] = val;
			}
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"scan") == 0) {
		ad7490_scan(devno);
	}
	else
		return(CMD_PARAM_ERROR);

	if (rc != 0)
		printf("RC: %d\n",rc);

	return(CMD_SUCCESS);
}
