/*
 * at25df321.c:
 * The AT25DF321 is a 32Mbit serial firmware dataflash memory.
 *
 * The AT25DF321 is tied to the first chip select on the
 * blackfin (it can also be used for bootup when the BF537 is
 * in the appropriate boot mode).
 *
 * The AT25DF321 is a 4MByte device, broken up into different sector
 * ranges depending on what you want to do...
 *
 * For sector protection, there are 64 64K sectors.
 * For block erase using the 'D8' command, there are 64 64K sectors.
 * For block erase using the '52' command, there are 128 32K sectors.
 * For block erase using the '20' command, there are 1024 4K sectors.
 * The device page size is 256 bytes (for writes).
 */
#include "config.h"
#include "cdefBF537.h"
#include "genlib.h"
#include "ctype.h"
#include "cli.h"
#include "cpu.h"
#include "spi_bf537.h"
#include "spif.h"

unsigned char devId[3] = { 0x1f, 0x47, 0x00 };

#define READ_FIRST	/* This enables code in the sector erase function
					 * that will first read to see if the sector is
					 * already erased prior to issueing the erase command.
					 * Unfortunately, this also requires a 64K buffer.
					 */

static long at25df321_baud;

void
spifQinit(void)
{
	*pSPI_BAUD = (SYSCLKFREQ/(at25df321_baud*2));
	__builtin_bfin_ssync();

	/* Enable SPI interface as...
	 *  Master, MSB first, 
	 */
	*pSPI_CTL = (SPE | MSTR | CPOL | CPHA);
	__builtin_bfin_ssync();
}

int
spifInit(void)
{
	char *env;

	/* The AT25DF321 is on SPI CS1 line, so enable that line.
	 */
	*pSPI_FLG |= FLS1;
	__builtin_bfin_ssync();

	if ((env = getenv("SPIFBAUD")) != (char *)0)
		at25df321_baud = strtol(env,0,0);
	else
		at25df321_baud = SPIFBAUD;

	spifQinit();

	*pSPI_STAT = (MODF | TXE | RBSY | TXCOL);
	__builtin_bfin_ssync();

	return(0);
}


int
spifWaitForReady(void)
{
	unsigned short stat;
	volatile int timeout = 1000000;

	while(timeout > 0) {
		stat = spifReadStatus(0);	
		if ((stat & BSY) == 0)
			break;
		//monDelay(10);
		timeout--;
	}
	if (timeout == 0)
		return(-1);
	return(0);
}

int
spifId(verbose)
{
	int i, rc;
	volatile unsigned short tmp;
	volatile unsigned char buf[3];

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	/* Write id request command:
	 */
	*pSPI_TDBR = 0x009f;	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE */

	SPI_FINWAIT();

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;

	SPI_FINWAIT();

	for(i=0;i<3;i++) {			/* Retrieve 3 bytes from device. */
		*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
		SPI_RXSWAIT();			/* Wait for receive buffer full */
		buf[i] = (unsigned char)*pSPI_RDBR;
	}

	SPI_FINWAIT();
	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	if (memcmp((char *)buf,(char *)devId,3) != 0) {
		if (verbose)
			printf("Unexpected SPI-Flash Id: ");
		rc = -1;
	}
	else {
		if (verbose)
			printf("SPI-Flash Id: 0x");
		rc = 0;
	}

	if (verbose) {
		for(i=0;i<3;i++) 
			printf("%02x",buf[i]);
		printf("\n");
	}
	return(rc);
}

unsigned short
spifReadStatus(int verbose)
{
	volatile unsigned short tmp;
	volatile unsigned short stat;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	/* Write id request command:
	 */
	*pSPI_TDBR = 0x0005;	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE */
	SPI_FINWAIT();

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;
	SPI_FINWAIT();

	*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
	SPI_RXSWAIT();			/* Wait for receive buffer full */
	stat = *pSPI_RDBR;

	SPI_FINWAIT();
	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	if (verbose) {
		printf("stat: 0x%02x\n",stat);
		if (stat & SPRL) printf("  Sector protection regs locked\n");
		if (stat & EPE)  printf("  Erase or program error\n");
		if (stat & WPP)  printf("  *WP pin is deasserted\n");
		switch(stat & SWPS) {
			case SWPS00:
				printf("  All sectors unprotected\n");
				break;
			case SWPS01:
				printf("  Some sectors unprotected\n");
				break;
			case SWPS11:
				printf("  All sectors protected\n");
				break;
		}
		if (stat & WEL) printf("  Write enable latch enabled\n");
		if (stat & BSY) printf("  Device is busy\n");
		printf("  Baud: %d (reg=0x%x)\n",
			(SYSCLKFREQ/(*pSPI_BAUD * 2)),*pSPI_BAUD);
	}

	return(stat);
}

int
spifWriteEnable(void)
{
	volatile unsigned short tmp;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	*pSPI_TDBR = 0x0006;	/* WRITE-ENABLE command. */
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();				/* Wait for completion. */

	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
spifWriteDisable(void)
{
	volatile unsigned short tmp;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	*pSPI_TDBR = 0x0004;	/* WRITE-DISABLE command. */
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();				/* Wait for completion. */

	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
spifBlockErase(int bsize, long addr)
{
	int i, rc;
	unsigned char buf[4];
	volatile unsigned short tmp;
#ifdef READ_FIRST
	unsigned long *lp, *end;
	static unsigned long block[0x10000/4];
#endif

	switch(bsize) {
		case 0x1000:
			buf[0] = BLKSZ_4K;
			break;
		case 0x8000:
			buf[0] = BLKSZ_32K;
			break;
		case 0x10000:
			buf[0] = BLKSZ_64K;
			break;
		default:
			return(-1);
	}
	rc = bsize;

#ifdef READ_FIRST
	/* Before we actually do the erase, see if the block is
	 * already erased.  If it is, then just return.  
	 * NOTE: I implemented this and tried it, and its not much
	 * faster; so, for now it is disabled.
	 */
	if (spifReadBlock(addr,(char *)block,bsize) == bsize) {
		lp = block; end = &block[bsize/4];
		while(lp < end) {
			if (*lp != 0xffffffff)
				break;
			lp++;
		}
		if (lp == end) {
			return(rc);
		}
	}
#endif		

	spifInit();

	SPI_ENABLE();
	__builtin_bfin_ssync();

	SPI_WRITE_MODE();
	__builtin_bfin_ssync();

	*pSPI_FLG &= FLG1;
	__builtin_bfin_ssync();

	buf[1] = ((addr & 0xff0000) >> 16);
	buf[2] = ((addr & 0xff00) >> 8);
	buf[3] = ((addr & 0xff));

	for(i=0;i<4;i++) {
		*pSPI_TDBR = buf[i];	// Load xmit register
		__builtin_bfin_ssync();
		tmp = *pSPI_RDBR;		// Do dummy read to kick off the transfer.
		SPI_TXSWAIT();
	}

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;
	__builtin_bfin_ssync();

	SPI_DISABLE();
	__builtin_bfin_ssync();

	spifWaitForReady();
		
	return(rc);
}

int
spifChipErase(void)
{
	volatile unsigned short tmp;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	*pSPI_TDBR = 0x00C7;	/* CHIP-ERASE command (C7 or 60 could be used). */	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
spifGlobalUnprotect(void)
{
	volatile unsigned short tmp;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	*pSPI_TDBR = 0x0001;	/* WRITE-STATUS-REG command. */	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();

	*pSPI_TDBR = 0x0000;	/* Global unprotect. */
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
spifGlobalProtect(void)
{
	volatile unsigned short tmp;

	spifInit();

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	*pSPI_TDBR = 0x0001;	/* WRITE-STATUS-REG command. */	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();

	*pSPI_TDBR = 0x8000;	/* Global protect. */
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE. */
	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
spifReadBlock(unsigned long addr,char *data,int len)
{
	int i;
	volatile unsigned short tmp;
	unsigned char buf[5];

	spifInit();

	SPI_ENABLE();
	__builtin_bfin_ssync();

	SPI_WRITE_MODE();
	__builtin_bfin_ssync();

	*pSPI_FLG &= FLG1;
	__builtin_bfin_ssync();

	buf[0] = 0x0b;
	buf[1] = ((addr & 0xff0000) >> 16);
	buf[2] = ((addr & 0xff00) >> 8);
	buf[3] = ((addr & 0xff));
	buf[4] = 0x00;				// Don't care (only needed with 0x0b opcode)

	for(i=0;i<5;i++) {
		*pSPI_TDBR = buf[i];	// Load xmit register
		__builtin_bfin_ssync();
		tmp = *pSPI_RDBR;		// Do dummy read to kick off the transfer.
		SPI_FINWAIT();
	}

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;
	SPI_FINWAIT();

	for(i=0;i<len;i++) {	
		*pSPI_TDBR = 0;	
		SPI_RXSWAIT();
		data[i] = *pSPI_RDBR;
	}

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;
	__builtin_bfin_ssync();

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(len);
}

static int 
spifWriteChunk(unsigned long addr, char *data, int len)
{
	int i;
	volatile unsigned short tmp;
	unsigned char buf[5];

	if (addr & 0xff000000) {
		printf("SPIF: addr out of range\n");
		return(-1);
	}
	if (len > 256) {
		printf("SPIF: chunk too big\n");
		return(-1);
	}

	if (spifWaitForReady() == -1)
		return(-1);
	if (spifWriteEnable() == -1)
		return(-1);

	spifInit();

	/* Send opcode and 3-byte address:
	 */
	buf[0] = 0x02;
	buf[1] = ((addr & 0xff0000) >> 16);
	buf[2] = ((addr & 0xff00) >> 8);
	buf[3] = ((addr & 0xff));

	SPI_ENABLE();
	__builtin_bfin_ssync();

	SPI_WRITE_MODE();
	__builtin_bfin_ssync();

	*pSPI_FLG &= FLG1;
	__builtin_bfin_ssync();

	/* Send command and address:
	 */
	for(i=0;i<4;i++) {
		*pSPI_TDBR = buf[i];	// Load xmit register
		__builtin_bfin_ssync();
		tmp = *pSPI_RDBR;		// Do dummy read to kick off the transfer.
		SPI_TXSWAIT();
	}

	/* Send data:
	 */
	for(i=0;i<len;i++) {
		*pSPI_TDBR = data[i];	// Load xmit register
		__builtin_bfin_ssync();
		tmp = *pSPI_RDBR;		// Do dummy read to kick off the transfer.
		SPI_TXSWAIT();
	}

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG1;
	__builtin_bfin_ssync();

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(len);
}

int 
spifWriteBlock(unsigned long addr, char *data, int len)
{
	int rc = len;

	/* If the starting address doesn't align on a page boundary, then we
	 * must make sure that we only write from the starting address to the
	 * next page boundary; otherwise the internal buffering of the device
	 * will cause the data to wrap (refer to the datasheet for full details).
	 */
	if (addr & 0xff) {
		char done = 0;
		int len1 = 256 - (addr & 0xff);

		//printf("Addr: 0x%x, len=%d, len1=%d\n",addr,len,len1);
		if (len1 >= len) {
			len1 = len;
			done = 1;
		}

		if (spifWriteChunk(addr,data,len1) == -1)
			rc = -1;

		if (done)
			return(rc);

		addr += len1;
		len -= len1;
		data += len1;
		//printf("addr: 0x%x, len=%d\n",addr,len);
	}

	if (len <= 256) {
		if (spifWriteChunk(addr,data,len) == -1)
			rc = -1;
	}
	else {
		while(len > 256) {
			if (spifWriteChunk(addr,data,256) == -1) {
				rc = -1;
				break;
			}
			addr += 256;
			data += 256;
			len -= 256;
		}
		if (len > 0) {
			if (spifWriteChunk(addr,data,len) == -1)
				rc = -1;
		}
	}
	return(rc);
}
