/* etherdev.c:
 * Hooks needed between uMon and BF537's ethernet MAC.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */

#include "config.h"
#if INCLUDE_ETHERNET
#include "genlib.h"
#include "stddefs.h"
#include "warmstart.h"
#include "ether.h"
#include "phy.h"
#include "cpu.h"
#include "cdefBF537.h"
#include "timer.h"
#include "cli.h"
#include "spif.h"
#include "spifmap.h"

int getPhyAdd(void);

#define PHY_VERBOSE
#define AUTO_RX_CKSUM		/* If defined, Blackfin will compute incoming UDP
							 * and TCP checksums (need to add code to take
							 * advantage of this).
							 */
#define DEFAULT_PHYADDR 0
#define MDCFREQ			2000000

#define RXTOT			4
#define TBUFSIZE		1560 
#define RBUFSIZE		1560 

int phy_mode(int device, int *mode);

/* Structure from HW Reference Manual...
 */
struct dma_descriptor {
	struct dma_descriptor		*next;
	ulong						*startaddr;
	ushort						config;
};

/* Transmit & receive data buffers:
 * TODO: consider using L1-resident ethernet buffers.
 */

int rxidx;


#ifdef TX_STAT_BASE
vulong *pTxStat = (vulong *)TX_STAT_BASE;
vulong *pTxData = (vulong *)TX_DATA_BASE;
vulong (*pRxStat)[2] = RX_STAT_BASE;
vulong (*pRxData)[(RBUFSIZE/sizeof(long))] = RX_DATA_BASE;
#else
vulong tx_stat;
vulong tx_data[(TBUFSIZE/sizeof(long))];
vulong rx_stat[RXTOT][2];
vulong rx_data[RXTOT][(RBUFSIZE/sizeof(long))];
vulong *pTxStat = &tx_stat;
vulong *pTxData = tx_data;
vulong (*pRxStat)[2] = rx_stat;
vulong (*pRxData)[(RBUFSIZE/sizeof(long))] = rx_data;
#endif

#ifdef TX_DMA_DESCRIPTOR_BASE
struct dma_descriptor *txd = (struct dma_descriptor *)TX_DMA_DESCRIPTOR_BASE;
#else
struct dma_descriptor txd[2];
#endif

#ifdef RX_DMA_DESCRIPTOR_BASE
struct dma_descriptor *rxd = (struct dma_descriptor *)RX_DMA_DESCRIPTOR_BASE;
#else
struct dma_descriptor rxd[RXTOT*2];
#endif

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	int mode;
	ushort ctrl;

	if (StateOfMonitor != INITIALIZE)
		return;

	phy_read(0,PHY_CTRL,&ctrl);
	ctrl &= ~PHY_CTRL_PWRDN;		/* Pull PHY out of power-down mode */
	ctrl &= ~PHY_CTRL_LBK;			/* Pull PHY out of loopback mode */
	phy_write(0,PHY_CTRL,ctrl); 

	phy_mode(0,&mode);
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_DRIVER_DEBUG)
		printf("Phy mode: 0x%02x\n",mode);
#endif

	/* If the PHY doesn't default to auto-negotiate (mode==7), then
	 * reset it here and then start it up later...
	 */
	if (mode != 7)
		phy_reset(0);

	*pEMAC_OPMODE = 0x00000000;
	*pDMA1_CONFIG = 0x0000;
	*pDMA2_CONFIG = 0x0000;
}

/*
 * eninit():
 * Initialize the PHY and MAC.
 * These steps are based on the section titled "Programming Model"
 * in chapter 8 (Ethernet MAC) of the hardware reference manual.
 *
 * Return 0 if successful; else -1.
 */
static int
eninit(void)
{
	static char beenhere;
	int		mode;
	ushort	sval, phyreg;
	ulong	lval, tmp;

	tmp = *pPORTH_FER;			/* Read it once (bug work-around) */
	*pPORTH_FER = 0xffff;		/* GPIO pin multiplexing (write it twice) */
	*pPORTH_FER = 0xffff;
	
	/* Load MAC address into device...
	 */
	memcpy((char *)&lval,(char *)BinEnetAddr,4);
	*pEMAC_ADDRLO = lval;
	memcpy((char *)&sval,(char *)BinEnetAddr+4,2);
	*pEMAC_ADDRHI = sval;

	/* Initialize SYSCTL:
	 */
#ifdef AUTO_RX_CKSUM
	*pEMAC_SYSCTL = (((SYSCLKFREQ/MDCFREQ)/2 - 1) << 8) | (RXDWA | RXCKS);
#else
	*pEMAC_SYSCTL = (((SYSCLKFREQ/MDCFREQ)/2 - 1) << 8) | RXDWA;
#endif

	/* PHY setup:
	 * The PHY is an SMSC LAN83C185 (on the board used for this port),
	 * the PHY ID should be 0x0007c0aX (X = revision).
	 */
	if (phy_id(0,&lval) == -1) {
		printf("phy access failure\n");
		return(-1);
	}

	/* Can't get proper link status out of the PHY immediately after reset.
	 * So, always invoke auto-negotiation...
	 */
	phy_mode(0,&mode);
	if (mode != 7)
		phy_autoneg(0);

	/* Setup MAC:
	 * If PHY is configured full duplex, then we wanna setup the
	 * MAC for full duplex...
	 */
	phy_read(0,PHY_CTRL,&phyreg);
	if (phyreg & PHY_CTRL_FDX)
		*pEMAC_OPMODE = ASTP | PSF | DRO | FDMODE;
	else
		*pEMAC_OPMODE = ASTP | PSF | DRO ;

	/* Establish/reset management counters:
	 */
	*pEMAC_MMC_CTL = MMCE | CCOR | RSTC;
	*pEMAC_MMC_CTL = MMCE | CCOR;
	
	/* Establish transmit and receive descriptors:
	 */
	*pTxStat = 0;
	txd[0].next = &txd[1];
	txd[0].startaddr = (ulong *)pTxData;
	txd[0].config = DMAEN | NDSIZE_5 | FLOW_LARGE | WDSIZE_32;
	txd[1].next = &txd[0];
	txd[1].startaddr = (ulong *)pTxStat;
	txd[1].config = DMAEN | DI_EN | WNR | WDSIZE_32;
	
	for(rxidx=0;rxidx<RXTOT;rxidx++) {
		pRxStat[rxidx][0] = 0;
		pRxStat[rxidx][1] = 0;
	}
	for(rxidx=0;rxidx<RXTOT*2;rxidx+=2) {
		rxd[rxidx].next = &rxd[rxidx+1];
		rxd[rxidx].startaddr = (ulong *)pRxData[rxidx/2];
		rxd[rxidx].config = WNR | DMAEN | NDSIZE_5 | FLOW_LARGE | WDSIZE_32;
		rxd[rxidx+1].next = &rxd[rxidx+2];
		rxd[rxidx+1].startaddr = (ulong *)&pRxStat[rxidx/2][0];
		rxd[rxidx+1].config =
			WNR | DMAEN | DI_EN | NDSIZE_5 | FLOW_LARGE | WDSIZE_32;
	}
	rxd[(RXTOT*2)-1].next = &rxd[0];
	rxidx = 0;

	/* Init DMA:
	 * DMA_1 is for ethernet receive
	 * DMA_2 is for ethernet transmit
	 */
	*pDMA1_X_COUNT  = 0;
	*pDMA1_X_MODIFY = 4;
	*pDMA1_Y_COUNT  = 0;
	*pDMA1_Y_MODIFY = 0;
	*pDMA2_X_COUNT  = 0;
	*pDMA2_X_MODIFY = 4;
	*pDMA2_Y_COUNT  = 0;
	*pDMA2_Y_MODIFY = 0;
	*pDMA1_NEXT_DESC_PTR = rxd;
	*pDMA1_CONFIG = DMAEN | WNR | NDSIZE_5 | FLOW_LARGE | WDSIZE_32;

	/* Once everything else is initialized, enable xmt & rcv...
	 */
	*pEMAC_OPMODE |= (TE | RE);
	beenhere = 1;
	return (0);
}

int
EtherdevStartup(int verbose)
{
	static char beenhere;

	if (StateOfMonitor != INITIALIZE)
		return(0);

	/* On the very first pass through this function we don't bring up
	 * the MAC.  Later in the initialization, polletherdev() will be
	 * called and it will detect the link-up status and automatically
	 * call eninit() to invoke auto-negotiation.
	 */
	if (!beenhere) {
		beenhere = 1;
		return(0);
	}

	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller and return the value returned by
	 * eninit().
	 */
	return(eninit());
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	*pEMAC_OPMODE &= ~PR;
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	*pEMAC_OPMODE |= PR;
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	*pEMAC_OPMODE |= DBF;
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	*pEMAC_OPMODE &= ~DBF;
}

static	uchar testmsg[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0x08, 0x00, 't', 'e', 's', 't'
};

/* xxxtestfailure():
 * These functions provide the user with the option to enter an endless
 * loop that will help debug the interface if something fails in the
 * self test...
 */
void
phytestfailure(void)
{
	ulong	phyid;

	printf("loop forever on phy access (y or n)?");
	if (getchar() == 'y') {
		while(1)
			phy_id(0,&phyid);
	}
}

void
lbktestfailure(void)
{
	uchar	*msg;

	printf("loop forever on packet send (y or n)?");
	if (getchar() == 'y') {
		while(1) {
			msg = getXmitBuffer();
			memcpy((char *)msg,(char *)testmsg,sizeof(testmsg));
			sendBuffer(64);
		}
	}
}

/* 
 * enselftest():
 *	Do a simple read of the PHY ID register and verify that it is
 *	as expected.  Then put the PHY in loopback and verify that what
 *	is transmitted is what is received.
 *	Return 1 if passed, else -1.
 */
int
enselftest(int verbose)
{
	char	*cp;
    struct	elapsed_tmr tmr;
	int		retval;
	uchar	*msg;
	ushort	phyreg;
	ulong	stat, phyid;

	retval = 1;
	if (verbose)
		printf("Self test... ");

	/* First read the PHY ID registers and verify valid device ID.
	 */
	if (phy_id(0,&phyid) == -1) {
		printf("phy access failed, using PHYADDR %d\n",getPhyAdd());
		phytestfailure();
		return(-1);
	}

	if (phyid != 0x1378e2) {
		printf("phy-id (0x%lx) failed\n",phyid);
		phytestfailure();
		return(-1);
	}

	/* Next, put the PHY in loopback and send a packet to see if we receive
	 * what we transmit...
	 */
	phy_read(0,PHY_CTRL,&phyreg);
	if (phy_write(0,PHY_CTRL,phyreg | PHY_CTRL_LBK) == -1) {
		if (verbose)
			printf("phy write failed\n");
		return(-1);
	}
	phy_read(0,PHY_CTRL,&phyreg);
	msg = getXmitBuffer();
	memcpy((char *)msg,(char *)testmsg,sizeof(testmsg));
	sendBuffer(64);

	/* Wait for the looped back message... */

    startElapsedTimer(&tmr,2500);
	do {
    	if(msecElapsed(&tmr)) {
			printf("LBK packet not rcvd\n");
			lbktestfailure();
			return(-1);
		}
#ifdef AUTO_RX_CKSUM
		stat = pRxStat[rxidx][1];
#else
		stat = pRxStat[rxidx][0];
#endif

	} while((stat & RX_COMP) == 0);

	if (stat & RX_OK) {
		cp = (char *)pRxData[rxidx];
		cp += 2;
	}
	else {
		printf("RX_OK failed\n");
		lbktestfailure();
		return(-1);
	}

	*pDMA1_IRQ_STATUS |= (DMA_DONE | DMA_ERR);
	pRxStat[rxidx][0] = 0;
	pRxStat[rxidx][1] = 0;

	if (++rxidx == RXTOT)
		rxidx = 0;

	if (memcmp(cp,(char *)testmsg,sizeof(testmsg))) {
		printf("lbk failed\n");
		lbktestfailure();
		retval = -1;
		goto done;
	}

	if (verbose)
		printf("passed\n");

	/* Put things back to normal... */
done:
	enreset();
	eninit();
	return(retval);
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	ushort phyreg;

	/* Read the register twice because the STAT bit is sticky,
	 * and we want to make sure that we are showing the current status.
	 */
	phy_read(0,PHY_STAT,&phyreg);
	phy_read(0,PHY_STAT,&phyreg);
	if (phyreg & PHY_STAT_LINKUP) {
		phy_read(0,PHY_CTRL,&phyreg);
		printf("PHY: %dMbps, %s-duplex\n",
			phyreg & PHY_CTRL_100MB ? 100 : 10,
			phyreg & PHY_CTRL_FDX ? "Full" : "Half");
	}
	else {
		printf("PHY: link is down\n");
	}
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  The monitor will never need more than one packet at
 * a time; hence, we just keep giving out the same pointer.
 * Note that for the BF537, the first two bytes of the buffer are used
 * by the DMA engine for the length of the packet; so this function actually
 * hands the requestor a pointer incremented by two from the base of tx_data.
 */
uchar *
getXmitBuffer(void)
{
	uchar *cp;
    struct elapsed_tmr tmr;

	/* Since we're only using one transmit buffer, we need to make sure
 	 * that the previous packet has been sent prior to giving the pointer
	 * up for re-use...
	 */
    startElapsedTimer(&tmr,5000);
	while(*pDMA2_IRQ_STATUS & DMA_RUN) {
    	if(msecElapsed(&tmr)) {
			printf("Wait-for-DMA_RUN timeout\n");
			break;
		}
	}
	cp = (uchar *)pTxData;
	cp += 2;
	return(cp);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 *
 * This driver uses register based DMA for ethernet transmit simply
 * because uMon only needs one transmit buffer; hence, there's really
 * no need for the added complexity of the descriptor model.
 * Note that DMA_2 is for ethernet transmit.
 */
int
sendBuffer(int length)
{
    struct elapsed_tmr tmr;
	int linkcheck(void);
	static char beenhere;
	int sentlen;
	ushort *sp;

	if (!beenhere) {
		beenhere = 1;
		if (linkcheck() == 0)
			return(0);
	}

	sp = (ushort *)pTxData;

	asm("ssync;");
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)(sp+1),length,ETHER_OUTGOING);
#endif

	*pTxStat = 0;
	asm("ssync;");
	*sp = length;
	asm("ssync;");
	*pDMA2_NEXT_DESC_PTR = txd;
	asm("ssync;");
	*pDMA2_CONFIG = DMAEN | NDSIZE_5 | FLOW_LARGE | WDSIZE_32;
	asm("ssync;");

    startElapsedTimer(&tmr,5000);
	while((*pTxStat & TX_COMP) == 0) {
		asm("ssync;");
    	if(msecElapsed(&tmr)) {
			printf("EMAC: TX_COMP timeout\n");
			return(0);
		}
	}
	asm("ssync;");
	if ((*pTxStat & TX_OK) == 0) 
		printf("EMAC txerr 0x%lx\n",*pTxStat);

	sentlen = (*pTxStat & TX_FRLEN) >> 16;
	
	asm("ssync;");
	if (sentlen < length) {
		printf("EMAC warning: sent count (%d) < send request (%d)\n",
			sentlen,length);
	}

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printf("  Sent %d bytes\n",sentlen);
#endif

	asm("ssync;");
	*pTxStat = 0;
	EtherXFRAMECnt++;
	return(0);
}

/* DisableEtherdev():
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * The VTC stores EVERYTHING to SPI flash.  The memory map of this device
 * is discussed in spifmap.h.  The MAC address is retrieved from the location
 * within SPI flash defined by SPIF_MAC_ADDRS_BASE.  It is assumed to be
 * ASCII data (just like it is normally stored in uMon NOR flash).
 */
char *
extGetEtherAdd(void)
{
	static unsigned char macstr[64];

	if (spifReadBlock(SPIF_MAC_ADDRS_BASE,(char *)macstr,sizeof(macstr)) < 0) {
		printf("extGetEtherAdd: spifReadBlock error\n");
		return((char *)0);
	}
	if (macstr[0] == 0xff)
		return(0);

	return((char *)macstr);
}

void
storeMac(int verbose)
{
	int yes, rc;
	char macascii[24], buf[6];

	if (spifReadBlock(SPIF_MAC_ADDRS_BASE,macascii,sizeof(macascii)) < 0) {
		printf("storeMac: spifReadBlock error\n");
		return;
	}
	if (macascii[0] != (char)0xff) {
		printf("MAC store abort: etheraddr[] contains data\n");
		return;
	}

	do {
		printf("Enter MAC address (xx:xx:xx:xx:xx:xx):\n");
		if (getline(macascii,sizeof(macascii),0) == 0)
			return;
	} while (EtherToBin((char *)macascii,(uchar *)buf) == -1);

	printf("Configuring '%s' as MAC address, ok?",macascii);
	yes =  askuser(" (y or n)");

	putchar('\n');
	if (!yes)
		return;
	
	spifWriteEnable();	
	spifGlobalUnprotect();	
	spifWriteEnable();	
	rc = spifWriteBlock(SPIF_MAC_ADDRS_BASE,macascii,strlen(macascii)+1);
	spifWriteEnable();	
	spifGlobalProtect();
	spifWriteDisable();	

	if (rc < 0) {
		printf("storeMac: spifWriteBlock error\n");
		return;
	}

	printf("MAC address burned into SPI flash\n");
}

int
linkcheck(void)
{
	ushort tmp;
	static ushort phystat;

	phy_read(0,PHY_STAT,&tmp);
	if ((tmp & PHY_STAT_LINKUP) && ((phystat & PHY_STAT_LINKUP) == 0)) {
		enreset();
		eninit();
		phystat = tmp;
	}
	else if (((tmp & PHY_STAT_LINKUP) == 0) && (phystat & PHY_STAT_LINKUP)) {
#if INCLUDE_ETHERVERBOSE
		if (EtherVerbose &  SHOW_OUTGOING)
			printf("\nEMAC: link down\n");
#endif
		phystat = tmp;
	}

	/* Return non-zero if link is up...
	 */
	return(phystat & PHY_STAT_LINKUP);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 *
 * NOTES:
 * 1. This function must be reentrant, because there are a few cases in
 *    processPACKET() where pollethernet() may be called.
 * 2. It should only process one packet per call.  This is important
 *    because if allowed to stay here to flush all available packets,
 *    it may starve the rest of the system (especially in cases of heavy
 *    network traffic).
 * 3. There are cases in the monitor's execution that may cause the
 *    polling polletherdev() to cease for several seconds.  Depending on
 *    network traffic, this may cause the input buffering mechanism on
 *    the ethernet device to overflow.  A robust polletherdev() function
 *    should support this gracefully (i.e. when the error is detected,
 *    attempt to pass all queued packets to processPACKET(), then do what
 *    is necessary to clear the error).
 */
int
polletherdev(void)
{
	static int filter;
	char *cp;
	ulong stat;
	int	pktcnt = 0;

#ifdef AUTO_RX_CKSUM
	stat = pRxStat[rxidx][1];
#else
	stat = pRxStat[rxidx][0];
#endif

	if (stat & RX_COMP) {
		if (stat & RX_OK) {
			cp = (char *)pRxData[rxidx];
			cp += 2;
			processPACKET((struct ether_header *)cp, (stat & RX_FRLEN));
			pktcnt++;
			filter = 256;
		}

		*pDMA1_IRQ_STATUS |= (DMA_DONE | DMA_ERR);
		pRxStat[rxidx][0] = 0;
		pRxStat[rxidx][1] = 0;

		if (++rxidx == RXTOT)
			rxidx = 0;
	}
	else {
		/* If there is an incoming packet, then there's no need to 
	 	 * check the link for state (it must be up)...
		 */
		if (filter == 0) {
			linkcheck();
			filter = 256;
		}
	}

	filter--;
	return(pktcnt);
}

/*************************************************************************
 * PHY interface:
 *************************************************************************
 */

/* phy_busy_poll():
 * Wait in loop (with timeout) until PHY isn't busy.
 */
int
phy_busy_poll(void)
{
    struct elapsed_tmr tmr, tmr1;

	/* Wait till previous MDC/MDIO transaction has completed.
     */
    startElapsedTimer(&tmr,5000);
	while(*(vushort *)EMAC_STAADD & STABUSY) {
    	if(msecElapsed(&tmr)) {
			printf("phy_busy_poll timeout\n");
			return(-1);
		}
    	startElapsedTimer(&tmr1,10);
    	while(!msecElapsed(&tmr1));
	}
	return(0);
}


/* getPhyAdd():
 * Provide a means to override the default PHY ADDR value...
 */
int
getPhyAdd(void)
{
	char *env;

	if ((env = getenv("PHYADD")) != 0)
		return(atoi(env));

	return(DEFAULT_PHYADDR);
}

/* phy_read():
 * Read an off-chip register in a PHY through the MDC/MDIO port:
 */
int
phy_read(int phy, int reg, unsigned short *value)
{
	if (phy_busy_poll() < 0)
		return(-1);

	*(vulong *)EMAC_STAADD =
		(SET_PHYAD(getPhyAdd()) | SET_REGAD(reg) | STABUSY);

	if (phy_busy_poll() < 0)
		return(-1);

	*value = (ushort) *(vulong *)EMAC_STADAT;
	return(0);
}


/* phy_write():
 * Write an off-chip register in a PHY through the MDC/MDIO port:
 */
int
phy_write(int phy, int reg, unsigned short value)
{
	if (phy_busy_poll() < 0)
		return(-1);

	*(vulong *)EMAC_STADAT = value;

	*(vulong *)EMAC_STAADD =
		(SET_PHYAD(getPhyAdd()) | SET_REGAD(reg) | STAOP | STABUSY);

	return(0);
}

int
phy_mode(int device, int *mode)
{
	ushort phyreg;

	if (phy_read(0,18,&phyreg) < 0)
		return(-1);

	*mode = (phyreg & 0x0070) >> 4;

	return(0);
}

int
phy_linkup(void)
{
	ushort stat;

	if (phy_read(0,1,&stat) < 0)
		return(-1);

	if (stat & PHY_STAT_LINKUP)
		return(1);
	return(0);
}


/* phy_reset():
 * Reset the phy.
 */
int
phy_reset(int phy)
{
	ushort ctrl;
    struct elapsed_tmr tmr;

	phy_read(phy,PHY_CTRL,&ctrl);
	ctrl |= PHY_CTRL_RST;
	phy_write(phy,PHY_CTRL,ctrl);
    startElapsedTimer(&tmr,10);
	while(1) {
		if (phy_read(phy,PHY_CTRL,&ctrl) < 0)
			return(-1);
		if ((ctrl & PHY_CTRL_RST) == 0)
			break;
    	if (msecElapsed(&tmr)) {
			printf("phy reset timeout\n");
			return(-1);
		}
    }
	return(0);
}

/* phy_autoneg():
 * Go through the auto-negotiation process...
 */
int
phy_autoneg(int phy)
{
	volatile int retries;
	ushort ctrl, stat, autoneg, par;

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_DRIVER_DEBUG)
		printf("\nPHY Autonegotiate... ");
#endif

	retries = 1000;
	autoneg = PHY_ANAD_802_3;

	phy_read(phy,PHY_STAT,&stat);
	if((stat & PHY_STAT_ANABLE) == 0) {
		printf("phy does not support autoneg\n");
		return (-1);
	}
	
	/* Set appropriate advertisement bits based on this
	 * phy's capabilities...
	 */
	if (stat & PHY_STAT_100FDX)
		autoneg |= PHY_ANAD_100FDX;
	else if (stat & PHY_STAT_10FDX)
		autoneg |= PHY_ANAD_10FDX;
	else if (stat & PHY_STAT_100HDX)
		autoneg |= PHY_ANAD_100HDX;
	else if (stat & PHY_STAT_10HDX)
		autoneg |= PHY_ANAD_10HDX;
	phy_write(phy,PHY_ANAD, autoneg);

	phy_read(phy,PHY_CTRL,&ctrl);
	ctrl &= ~(PHY_CTRL_LBK | PHY_CTRL_PWRDN);
	ctrl &= ~(PHY_CTRL_ISOLATE | PHY_CTRL_COLTST);
	ctrl |= (PHY_CTRL_ANE);

	phy_write(phy,PHY_CTRL, ctrl);
	ctrl |= PHY_CTRL_ANRESTART; 
	phy_write(phy,PHY_CTRL, ctrl);

	do {
		phy_read(phy,PHY_STAT,&stat);
		if((stat & PHY_STAT_REMFAULT) != 0) {
			printf("phy: remote fault\n");
			return -1;
		}
		if (--retries <= 0) {
#if INCLUDE_ETHERVERBOSE
			if (EtherVerbose &  SHOW_DRIVER_DEBUG)
				printf("phy: AN timeout\n");
#endif
			return -1;
		}
	} while((stat & PHY_STAT_ANDONE) == 0);

	phy_read(phy,PHY_ANLPAR,&par);

	if((par & PHY_ANLP_REMFAULT) != 0) {
		printf("phy: remote fault\n");
		return -1;
	}

	phy_read(phy,PHY_CTRL,&ctrl);
	ctrl &= (~PHY_CTRL_FDX | PHY_CTRL_100MB);

	phy_read(phy,PHY_STAT,&stat);
	if ((stat & PHY_STAT_100FDX) && (par & PHY_ANLP_100FDX))
		ctrl |= (PHY_CTRL_100MB | PHY_CTRL_FDX);
	else if ((stat & PHY_STAT_10FDX) && (par & PHY_ANLP_10FDX))
		ctrl |= PHY_CTRL_FDX;
	else if ((stat & PHY_STAT_100HDX) && (par & PHY_ANLP_100HDX))
		ctrl |= PHY_CTRL_100MB;

	phy_write(phy,PHY_CTRL, ctrl);

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_DRIVER_DEBUG) {
		printf("%dMbps %s-duplex\n",
			ctrl & PHY_CTRL_100MB ? 100 : 10,
			ctrl & PHY_CTRL_FDX ? "Full" : "Half");
	}
#endif
	return 0;
}

/* phy_id():
 * Return the concatenation of the hi and lo phy ID registers (16 bits each)
 * as a single 32-bit value.
 */
int
phy_id(int phy, ulong *id)
{
	ushort idhi, idlo;

	if (phy_read(phy,PHY_IDHI,&idhi) < 0)
		return(-1);
	if (phy_read(phy,PHY_IDLO,&idlo) < 0)
		return(-1);
	*id = (((ulong)idhi << 16) | idlo);
	//printf("phyid: 0x%lx\n",*id);
	return(0);
}

char *
phyCmdHelp[] = {
	"Read/Write the PHY",
	"[-d:] {read regno | write regno value | reset | id}",
#if INCLUDE_VERBOSEHELP
	"Options:",
	" -d#   specify device # (default=0)",
#endif
	0,
};

int
phyCmd(int argc, char *argv[])
{
	int opt, device, regno;
	unsigned short value;
	unsigned long id;

	device = 0;
	while((opt=getopt(argc,argv,"d:")) != -1) {
		switch(opt) {
		case 'd':
			device = atoi(optarg);
			break;
		default:
			return(CMD_PARAM_ERROR);
		}
	}
	if (argc < optind+1)
		return(CMD_PARAM_ERROR);

	if (!strcmp(argv[optind],"read")) {
		if (argc != optind+2)
			return(CMD_PARAM_ERROR);
		regno = strtol(argv[optind+1],0,0);
		if (phy_read(device,regno,&value) < 0)
			return(CMD_FAILURE);
		printf("Reg %d: 0x%04x\n",regno,value);
	}
	else if (!strcmp(argv[optind],"write")) {
		if (argc != optind+3)
			return(CMD_PARAM_ERROR);
		regno = (unsigned short)strtol(argv[optind+1],0,0);
		value = (unsigned short)strtol(argv[optind+2],0,0);
		if (phy_write(device,regno,value) < 0)
			return(CMD_FAILURE);
	}
	else if (!strcmp(argv[optind],"reset")) {
		if (phy_reset(device) < 0)
			return(CMD_FAILURE);
	}
	else if (!strcmp(argv[optind],"id")) {
		if (phy_id(device,&id) < 0)
			return(CMD_FAILURE);
		printf("phyid: 0x%lx\n",id);
	}
	else 
		return(CMD_PARAM_ERROR);

	return(CMD_SUCCESS);
}

#endif
