//
// strtod.c
//
// Convert string to double 
//
// Copyright (C) 2002 Michael Ringgaard. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
// 1. Redistributions of source code must retain the above copyright 
//    notice, this list of conditions and the following disclaimer.  
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.  
// 3. Neither the name of the project nor the names of its contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
// SUCH DAMAGE.
// 

#include <ctype.h>
#include <float.h>
#include <stdlib.h>
#include <math.h>

double strtod(const char *str, char **endptr)
{
  double number;
  int exponent;
  int negative;
  char *p = (char *) str;
  double p10;
  int n;
  int num_digits;
  int num_decimals;

  // Skip leading whitespace
  while (isspace(*p)) p++;

  // Handle optional sign
  negative = 0;
  switch (*p) 
  {             
    case '-': negative = 1; // Fall through to increment position
    case '+': p++;
  }

  number = 0.;
  exponent = 0;
  num_digits = 0;
  num_decimals = 0;

  // Process string of digits
  while (isdigit(*p))
  {
    number = number * 10. + (*p - '0');
    p++;
    num_digits++;
  }

  // Process decimal part
  if (*p == '.') 
  {
    p++;

    while (isdigit(*p))
    {
      number = number * 10. + (*p - '0');
      p++;
      num_digits++;
      num_decimals++;
    }

    exponent -= num_decimals;
  }

  if (num_digits == 0)
  {
    //errno = ERANGE;
    return 0.0;
  }

  // Correct for sign
  if (negative) number = -number;

  // Process an exponent string
  if (*p == 'e' || *p == 'E') 
  {
    // Handle optional sign
    negative = 0;
    switch(*++p) 
    {   
      case '-': negative = 1;   // Fall through to increment pos
      case '+': p++;
    }

    // Process string of digits
    n = 0;
    while (isdigit(*p)) 
    {   
      n = n * 10 + (*p - '0');
      p++;
    }

    if (negative) 
      exponent -= n;
    else
      exponent += n;
  }

  if (exponent < DBL_MIN_EXP  || exponent > DBL_MAX_EXP)
  {
    //errno = ERANGE;
    return HUGE_VAL;
  }

  // Scale the result
  p10 = 10.;
  n = exponent;
  if (n < 0) n = -n;
  while (n) 
  {
    if (n & 1) 
    {
      if (exponent < 0)
        number /= p10;
      else
        number *= p10;
    }
    n >>= 1;
    p10 *= p10;
  }

  //if (number == HUGE_VAL) errno = ERANGE;
  if (endptr) *endptr = p;

  return number;
}

double atof(const char *str)
{
  return strtod(str, NULL);
}

////////////////////////////////////////////////////////
//
// The following code was not orginally part of the above

#include <string.h>

char *
ltoa(char *lbuf, long lval)
{
	unsigned long value;
	short size, minsize;
	char    *lbp;

	lbp = lbuf;
	minsize = 0;

	/* First determine how many ascii digits are needed. */
	value = abs(lval);
	size = 0;
	while (value > 0) {
		size++;
		value /= 10;
	}
	if (lval < 0)
		size++;
	if (minsize > size)
		size = minsize;
	lbp += size;

	/* Now build the string. */
	value = abs(lval);
	if (value == 0) {
		if (minsize == 0) {
			lbuf[0] = '0';
			size = 1;
		}else
			*--lbp = '0';
	}else  {
		while (value > 0) {
			*--lbp = (char)((value % 10) + '0');
			value /= 10;
		}
	}
	if (lval < 0)
		*--lbp = '-';
	lbuf[size] = 0;

	return(lbuf);
}

#define _FTOA_TOO_SMALL 99
#define _FTOA_TOO_LARGE 88

typedef union {
	long L;
	float F;
} LF_t;

char *
sprnfloat(char *buf, int bsize, float fpval, int trunc)
{
	LF_t	lf;
	char	*bp;
	short	exp2;
	long	mantissa, int_part, frac_part;

	if (fpval == 0.0) {
		strcpy(buf, "0.0");
		return buf;
	}
	lf.F = fpval;

	exp2 = (0xFF & (lf.L >> 23)) - 127;
	mantissa = (lf.L & 0xFFFFFF) | 0x800000;
	frac_part = 0;
	int_part = 0;

	if (exp2 >= 31) {
		strcpy(buf, "FPTOOBIG");
		return(buf);
	}else if (exp2 < -23) {
		strcpy(buf, "FPTOOSMALL");
		return(buf);
	}else if (exp2 >= 23)
		int_part = mantissa << (exp2 - 23);
	else if (exp2 >= 0) {
		int_part = mantissa >> (23 - exp2);
		frac_part = (mantissa << (exp2 + 1)) & 0xFFFFFF;
	}else /* if (exp2 < 0) */
		frac_part = (mantissa & 0xFFFFFF) >> -(exp2 + 1);

	bp = buf;

	if (lf.L < 0)
		*bp++ = '-';

	if (int_part == 0)
		*bp++ = '0';
	else {
#if 1
		ltoa(bp, int_part);
		while (*bp)
			bp++;
#else
		bp += long_to_dec(int_part,bp,buf+bsize-1,"");
#endif
	}
	*bp++ = '.';

	if (frac_part == 0)
		*bp++ = '0';
	else {
		char m, max;

		max = bsize - (bp - buf) - 1;
		if (max > trunc)
			max = trunc;
		/* print BCD */
		for (m = 0; m < max; m++) {
			/* frac_part *= 10; */
			frac_part = (frac_part << 3) + (frac_part << 1);

			*bp++ = (frac_part >> 24) + '0';
			frac_part &= 0xFFFFFF;
		}
		/* delete ending zeroes */
		for (--bp; bp[0] == '0' && bp[-1] != '.'; --bp)
			;
		++bp;
	}
	*bp = 0;

	return(buf);
}
